# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Charlie Vanaret, Benoit Pauwels, Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Design space
============

A design space is used to represent the optimization's unknowns,
a.k.a. design variables. A :class:`.DesignSpace` describes
this design space at a given state, in terms of names, sizes, types, bounds
and current values of the design variables. Variables can easily be added to
the :class:`.DesignSpace` using the :meth:`.DesignSpace.add_variable` method
or removed using the :meth:`.DesignSpace.remove_variable` method. We can also
filter the design variables using the :meth:`.DesignSpace.filter` method.
Getters and setters are also available to get or set the value of a given
variable property. Lastly, an instance of :class:`.DesignSpace` can be stored
in a txt or HDF file.
"""

from __future__ import division, unicode_literals

import logging
from copy import deepcopy

import h5py
from numpy import abs as np_abs
from numpy import (
    array,
    atleast_1d,
    complex128,
    concatenate,
    empty,
    equal,
    finfo,
    float64,
    genfromtxt,
    inf,
    int32,
    isnan,
    logical_or,
    mod,
    ndarray,
    ones,
    ones_like,
)
from numpy import round_ as np_round
from numpy import string_, vectorize, where, zeros_like
from six import string_types

from gemseo.algos.opt_result import OptimizationResult
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.hdf5 import get_hdf5_group
from gemseo.utils.py23_compat import string_array, strings_to_unicode_list

LOGGER = logging.getLogger(__name__)


class DesignSpace(object):
    """Class that describes the design space at a given state:

    the names/sizes/types/bounds of the variables and the initial solution of the
    optimization problem
    """

    FLOAT = "float"
    INTEGER = "integer"
    AVAILABLE_TYPES = [FLOAT, INTEGER]
    MINIMAL_FIELDS = ["name", "lower_bound", "upper_bound"]
    TABLE_NAMES = ["name", "lower_bound", "value", "upper_bound", "type"]

    DESIGN_SPACE_GROUP = "design_space"
    NAME_GROUP = "name"
    NAMES_GROUP = "names"
    LB_GROUP = "l_b"
    UB_GROUP = "u_b"
    TYPE_GROUP = "type"
    VAR_TYPE_GROUP = "var_type"
    VALUE_GROUP = "value"
    SIZE_GROUP = "size"
    # separator that denotes a vector's components
    SEP = "!"

    def __init__(self, hdf_file=None):
        """Constructor."""
        self.variables_names = []
        self.dimension = 0
        self.variables_sizes = {}
        self.variables_types = {}
        self.normalize = {}
        self._lower_bounds = {}
        self._upper_bounds = {}
        # These attributes are stored for faster computation of normalization
        # and unnormalization
        self._norm_factor = None
        self.__lower_bounds_array = None
        self.__upper_bounds_array = None
        self.__int_vars_indices = None
        self.__norm_data_is_computed = False
        self.__norm_inds = None
        self.__to_zero = None
        self.__bound_tol = 100.0 * finfo(float64).eps
        self._current_x = {}
        if hdf_file is not None:
            self.import_hdf(hdf_file)

    def remove_variable(self, name):
        """Remove a variable (and bounds and types) from the design space.

        :param name: name of the variable to remove
        """
        self.__norm_data_is_computed = False
        size = self.variables_sizes[name]
        self.dimension -= size
        self.variables_names.remove(name)
        del self.variables_sizes[name]
        del self.variables_types[name]
        del self.normalize[name]
        if name in self._lower_bounds:
            del self._lower_bounds[name]
        if name in self._upper_bounds:
            del self._upper_bounds[name]
        if name in self._current_x:
            del self._current_x[name]

    def filter(self, keep_variables, copy=False):
        """Filter the design space to keep a sublist of variables.

        :param keep_variables: the list of variables to keep
        :type keep_variables: str of list(str)
        :param copy: if True then a copy of the design space is filtered, otherwise the
            design space itself is filtered
        :type copy: bool
        :returns: the filtered design space (or a copy)
        :rtype: DesignSpace
        """
        if isinstance(keep_variables, string_types):
            keep_variables = [keep_variables]
        design_space = deepcopy(self) if copy else self
        for name in deepcopy(self.variables_names):
            if name not in keep_variables:
                design_space.remove_variable(name)
        for name in keep_variables:
            if name not in self.variables_names:
                raise ValueError('Variable "' + str(name) + '" is not known')
        return design_space

    def filter_dim(self, variable, keep_dimensions):
        """Filters the design space to keep a sublist of dimensions for a given
        variable.

        :param variable: the variable
        :param keep_dimensions: the list of dimension to keep
        """
        self.__norm_data_is_computed = False
        removed_dimensions = list(
            set(range(self.variables_sizes[variable])) - set(keep_dimensions)
        )
        bad_dimensions = list(
            set(keep_dimensions) - set(range(self.variables_sizes[variable]))
        )
        size = len(removed_dimensions)
        self.dimension -= size
        self.variables_sizes[variable] -= size
        types = []
        for dimension in keep_dimensions:
            if dimension in bad_dimensions:
                self.remove_variable(variable)
                raise ValueError(
                    "Dimension "
                    + str(dimension)
                    + ' of variable "'
                    + str(variable)
                    + '" is not known'
                )
            types.append(self.variables_types[variable][dimension])
            self.variables_types[variable] = array(types)
        idx = keep_dimensions
        self.normalize[variable] = self.normalize[variable][idx]
        if variable in self._lower_bounds:
            self._lower_bounds[variable] = self._lower_bounds[variable][idx]
        if variable in self._upper_bounds:
            self._upper_bounds[variable] = self._upper_bounds[variable][idx]
        if variable in self._current_x:
            self._current_x[variable] = self._current_x[variable][idx]
        return self

    def add_variable(
        self, name, size=1, var_type=FLOAT, l_b=None, u_b=None, value=None
    ):
        """Add a variable to the design space.

        :param name: param size:  (Default value = 1)
        :param var_type: Default value = FLOAT)
        :param l_b: Default value = None)
        :param u_b: Default value = None)
        :param value: Default value = None)
        :param size:  (Default value = 1)
        """
        if name in self.variables_names:
            raise ValueError("Variable " + name + " already exists")
        if size <= 0 or int(size) != size:
            raise ValueError("The size of " + name + " should be a positive integer.")

        # name and size
        self.variables_names.append(name)
        self.dimension += size
        self.variables_sizes[name] = size
        # type
        self._add_type(name, size, var_type)

        # bounds
        self._add_bound(name, size, l_b, is_lower=True)
        self._add_bound(name, size, u_b, is_lower=False)
        self._check_variable_bounds(name)

        # normalization policy
        self._add_norm_policy(name)

        if value is not None:
            array_value = atleast_1d(value)
            self._check_value(array_value, name)
            if len(array_value) == 1 and size > 1:
                array_value = array_value * ones(size)
            self._current_x[name] = array_value
            self._check_current_x_value(name)

    def _add_type(self, name, size, var_type=None):
        """Add a type to a variable.

        :param name: name of the variable
        :param size: size of the variable
        :param var_type: type in self.AVAILABLE_TYPES,
            or None, which is then FLOAT by default
        """
        if isinstance(var_type, bytes):
            var_type = var_type.decode()
        if var_type is None:
            var_type = self.FLOAT
        if hasattr(var_type, "__iter__") and not isinstance(var_type, string_types):
            if len(var_type) != size:
                raise ValueError(
                    "The list of types for  variable "
                    + name
                    + " should be of size "
                    + str(size)
                )
            # a type for each component
            var_type = [
                v_type.decode() if isinstance(v_type, bytes) else v_type
                for v_type in var_type
            ]
            for v_type in var_type:
                if v_type not in self.AVAILABLE_TYPES:
                    msg = 'The type "{0}" of {1} is not known'.format(v_type, name)
                    raise ValueError(msg)
            self.variables_types[name] = array(var_type)

        else:
            # same type for all components
            if var_type not in self.AVAILABLE_TYPES:
                raise ValueError('Type "' + str(var_type) + '" is not known')
            var_type_array = array([var_type] * size)
            self.variables_types[name] = var_type_array

        self.__norm_data_is_computed = False

    def _add_norm_policy(self, name):
        """Adds a normalization policy to a variable. Unbounded variables are not
        normalized. Bounded variables (both from above and from below) are normalized.

        :param name: variable name
        """
        # Check that the variable is in the design space:
        if name not in self.variables_names:
            raise ValueError("Variable " + name + " is not in the design " + "space.")
        # Check that the variable size is set:
        size = self.get_size(name)
        if size is None:
            raise ValueError("The size of variable " + name + " is not set.")
        # Check that the variables types are set:
        variables_types = self.variables_types.get(name, None)
        if variables_types is None:
            raise ValueError(
                "The components types of variable " + name + " are not set."
            )
        # Set the normalization policy:
        normalize = empty(size)
        for i in range(size):
            var_type = variables_types[i]
            if var_type in (DesignSpace.INTEGER, DesignSpace.FLOAT):
                if (
                    self._lower_bounds[name][i] == -inf
                    or self._upper_bounds[name][i] == inf
                ):
                    # Unbounded variables are not normalized:
                    normalize[i] = False
                elif self._lower_bounds[name][i] == self._upper_bounds[name][i]:
                    # Constant variables are not normalized:
                    normalize[i] = False
                else:
                    normalize[i] = True
            else:
                msg = "The normalization policy for type {0}" + " is not implemented."
                raise ValueError(msg.format(var_type))
        self.normalize[name] = normalize

    @staticmethod
    def __is_integer(value):
        """Checks that all values are integers."""
        are_none = equal(value, None)
        are_int = equal(mod(value.astype("f"), 1), 0)
        return logical_or(are_none, are_int)

    @staticmethod
    def __is_numeric(value):
        """Checks that a value is numeric."""
        res = (value is None) or hasattr(value, "real")
        try:
            if not res:
                float(value)
            return True
        except TypeError:
            return False

    @staticmethod
    def __isnot_nan(value):
        """Checks that a value is not nan."""
        return (value is None) or ~isnan(value)

    def _check_value(self, value, name):
        """Checks that a variable value is valid.

        :param value: a numpy array
        """
        all_indices = set(range(len(value)))
        # OK if the variable value is one-dimensional
        if len(value.shape) > 1:
            raise ValueError(
                "Value "
                + str(value)
                + " of variable "
                + str(name)
                + " has dimension greater than 1"
                + "while a float or a 1d iterable object "
                + "(array, list, tuple, ...) "
                + "while a scalar was expected."
            )

        # OK if all components are None
        if all(equal(value, None)):
            return True

        test = vectorize(self.__is_numeric)(value)
        indices = all_indices - set(list(where(test)[0]))
        for idx in indices:
            raise ValueError(
                "Value "
                + str(value[idx])
                + " of variable "
                + str(name)
                + " is not numerizable."
            )

        test = vectorize(self.__isnot_nan)(value)
        indices = all_indices - set(list(where(test)[0]))
        for idx in indices:
            raise ValueError(
                "Value " + str(value[idx]) + " of variable " + str(name) + " is nan."
            )

        # Check if some components of an integer variable are not integer.
        if self.variables_types[name][0] == self.INTEGER:
            indices = all_indices - set(list(where(self.__is_integer(value))[0]))
            for idx in indices:
                raise ValueError(
                    "Component value "
                    + str(value[idx])
                    + " of variable "
                    + str(name)
                    + " is not an integer "
                    + "while variable is of type integer"
                    + "(index: "
                    + str(idx)
                    + ")."
                )

    def _add_bound(self, name, size, bound, is_lower=True):
        """Add a lower or upper bound to a variable.

        :param name: name of the variable
        :param bound: lower or upper bound (array)
        :param size: size of the variable
        :param is_lower: if True, bound is a lower bound
        """
        self.__norm_data_is_computed = False
        if is_lower:
            bound_dict = self._lower_bounds
        else:
            bound_dict = self._upper_bounds
        infinity = inf * ones(size)
        if bound is None:
            if is_lower:
                bound_array = -infinity
            else:
                bound_array = infinity
            bound_dict.update({name: bound_array})
            return
        bound_array = atleast_1d(bound)
        self._check_value(bound_array, name)
        if hasattr(bound, "__iter__"):
            # iterable structure
            if len(bound_array) != size:
                bound_str = "lower" if is_lower else "upper"
                raise ValueError(
                    "The "
                    + bound_str
                    + " bounds of "
                    + name
                    + " should be of size "
                    + str(size)
                )
            if is_lower:
                bound_array = where(equal(bound_array, None), -infinity, bound_array)
            else:
                bound_array = where(equal(bound_array, None), infinity, bound_array)
            bound_dict.update({name: bound_array})
        else:
            # scalar: same lower bound for all components
            bound_array = bound * ones(size)
            bound_dict.update({name: bound_array})
        return

    def _check_variable_bounds(self, name):
        """Check that the bounds are compatible and are the same size.

        :param name: name of the variable
        """
        l_b = self._lower_bounds.get(name, None)
        u_b = self._upper_bounds.get(name, None)
        inds = where(u_b < l_b)[0]
        if inds.size != 0:
            raise ValueError(
                "The bounds of variable "
                + name
                + str(inds)
                + " are not valid : "
                + str(l_b[inds])
                + "!<"
                + str(u_b[inds])
            )

    def _check_current_x_value(self, name):
        """Check that the current x values are between bounds.

        :param name: name of the variable
        """
        l_b = self._lower_bounds.get(name, None)
        u_b = self._upper_bounds.get(name, None)
        c_x = self._current_x.get(name, None)
        not_none = ~equal(c_x, None)
        inds = where(
            logical_or(
                c_x[not_none] < l_b[not_none] - 1e-14,
                c_x[not_none] > u_b[not_none] + 1e-14,
            )
        )[0]
        for idx in inds:
            raise ValueError(
                "The current value of variable "
                + name
                + "!"
                + str(c_x[idx])
                + " is not between "
                + "the lower bound "
                + str(l_b[idx])
                + " and "
                + "the upper bound "
                + str(u_b[idx])
            )

    def has_current_x(self):
        """Tests if current_x is defined.

        :returns: True if current_x is defined
        """
        if self._current_x is None or len(self._current_x) != len(self.variables_names):
            return False
        for val in self._current_x.values():
            if val is not None:
                return True
        return False

    def check(self):
        """Check the state of the design space."""
        if not self.variables_names:
            raise ValueError("Design space is empty !")

        for name in self.variables_names:
            self._check_variable_bounds(name)

        if self.has_current_x():
            self._check_current_x()

    def check_membership(self, x_vect, variables_names=None):
        """Checks whether the input variables satisfy the design space requirements.

        :param x_vect: design variables
        :type x_vect: dict or array
        :param variables_names: names of the variables to be checked
        """
        # Convert the input vector into a dictionary if necessary:
        if isinstance(x_vect, dict):
            x_dict = x_vect
        elif isinstance(x_vect, ndarray):
            if x_vect.size != self.dimension:
                raise ValueError(
                    "The dimension of the input array ("
                    + str(x_vect.size)
                    + ") should be "
                    + str(self.dimension)
                    + "."
                )
            x_dict = self.array_to_dict(x_vect)
        else:
            raise TypeError(
                "The input vector should be an array or a "
                + "dictionary. Got "
                + str(type(x_vect))
                + " instead."
            )
        # Check the membership of the input vector to the design space:
        variables_names = variables_names or self.variables_names
        for name in variables_names:
            if x_dict[name] is None:
                continue
            size = self.variables_sizes[name]
            l_b = self._lower_bounds.get(name, None)
            u_b = self._upper_bounds.get(name, None)

            if atleast_1d(x_dict[name]).size != size:
                raise ValueError(
                    "The component "
                    + name
                    + " of the given"
                    + " array should have size "
                    + str(size)
                    + "."
                )
            for i in range(size):
                x_real = atleast_1d(x_dict[name])[i].real
                if l_b is not None and x_real < l_b[i] - self.__bound_tol:
                    msg = "The component " + name + self.SEP + str(i)
                    msg += " of the given array (" + str(x_real) + ") is "
                    msg += "lower than the lower bound (" + str(l_b[i]) + ")"
                    msg += " by {:.1e}.".format(l_b[i] - x_real)
                    raise ValueError(msg)
                if u_b is not None and u_b[i] + self.__bound_tol < x_real:
                    msg = "The component " + name + self.SEP + str(i)
                    msg += " of the given array (" + str(x_real) + ") is "
                    msg += "greater than the upper bound ("
                    msg += str(u_b[i]) + ")"
                    msg += " by {:.1e}.".format(x_real - u_b[i])
                    raise ValueError(msg)
                if (
                    self.variables_types[name][0] == self.INTEGER
                ) and not self.__is_integer(x_real):
                    msg = "The component " + name + self.SEP + str(i)
                    msg += " of the given array is not an integer "
                    msg += " while variable is of type integer !"
                    msg += " value = " + str(x_real)
                    raise ValueError(msg)

    def get_active_bounds(self, x_vec=None, tol=1e-8):
        """Determine which bound constraints of the current point are active.

        :param x_vec: the point at which we check the bounds
        :param tol: tolerance of comparison of a scalar with a bound
            (Default value = 1e-8)
        """
        if x_vec is None:
            x_dict = self._current_x
            self.check_membership(self.get_current_x())
        elif isinstance(x_vec, ndarray):
            x_dict = self.array_to_dict(x_vec)
        elif isinstance(x_vec, dict):
            x_dict = x_vec
        else:
            raise TypeError(
                "Expected dict or array for x_vec argument,"
                + " got "
                + str(type(x_vec))
            )

        active_l_b = {}
        active_u_b = {}
        for name in self.variables_names:
            l_b = self._lower_bounds.get(name)
            l_b = where(equal(l_b, None), -inf, l_b)
            u_b = self._upper_bounds.get(name)
            u_b = where(equal(u_b, None), inf, u_b)
            x_vec_i = x_dict[name]
            # lower bound saturated
            albi = where(np_abs(x_vec_i - l_b) <= tol, True, False)
            active_l_b[name] = albi
            # upper bound saturated
            aubi = where(np_abs(x_vec_i - u_b) <= tol, True, False)
            active_u_b[name] = aubi
        return active_l_b, active_u_b

    def _check_current_x(self, variables_names=None):
        """Checks the current point.

        :param variables_names: Default value = None)
        """
        if sorted(set(self.variables_names)) != sorted(set(self._current_x.keys())):
            raise ValueError(
                "Expected current_x variables :"
                + str(self.variables_names)
                + ", got "
                + str(list(self._current_x.keys()))
            )
        self.check_membership(self._current_x, variables_names)

    def get_current_x(self, variables_names=None):
        """Gets the current point in the design space.

        :param variables_names: names of the required variables, optional
        :type variables_names: list(str)
        :returns: the x vector as array
        :rtype: ndarray
        """
        try:
            x_arr = self.dict_to_array(self._current_x, all_var_list=variables_names)
            return x_arr
        except KeyError as err:
            raise KeyError("DesignSpace has no current_x for " + err.args[0])

    def get_indexed_var_name(self, variable_name):
        """Return a list of the variables names with their indices such as.

        [x!0,x!1,y,z!0,z!1]

        :param variable_name: name of the variable
        :type variable_name: str
        :returns: names of the variable components
        :rtype: list(str)
        """
        size = self.variables_sizes[variable_name]
        if size == 1:
            return variable_name
        return [variable_name + self.SEP + str(i) for i in range(size)]

    def get_indexed_variables_names(self):
        """Return a list of the variables names with their indices such as.

        [x!0,x!1,y,z!0,z!1]

        :returns: names of all the variables components
        :rtype: list(str)
        """
        var_ind_names = []
        for var in self.variables_names:
            vnames = self.get_indexed_var_name(var)
            if isinstance(vnames, string_types):
                var_ind_names.append(vnames)
            else:
                var_ind_names += vnames
        return var_ind_names

    def get_variables_indexes(self, variables_names):
        """Return the indexes of a design array corresponding to the variables names.

        :param variables_names: names of the variables
        :type variables_names: list(str)
        :return: indexes of a design array corresponding to the variables names
        :rtype: ndarray
        """
        indexes = list()
        index = 0
        for name in self.variables_names:
            var_size = self.get_size(name)
            if name in variables_names:
                indexes.extend(range(index, index + var_size))
            index += var_size
        return array(indexes)

    def __update_normalization_vars(self):
        """Computes inner attributes used to compute the
        normalization/unnormalization."""

        self.__lower_bounds_array = self.get_lower_bounds()
        self.__upper_bounds_array = self.get_upper_bounds()
        self._norm_factor = self.__upper_bounds_array - self.__lower_bounds_array
        norm_array = self.dict_to_array(self.normalize)
        self.__norm_inds = where(norm_array)[0]
        # In case lb=ub
        self.__to_zero = where(self._norm_factor == 0.0)[0]

        var_ind_list = []
        for var in self.variables_names:
            # Store the mask of int variables
            to_one = self.variables_types[var] == self.INTEGER
            var_ind_list.append(to_one)
        self.__int_vars_indices = concatenate(var_ind_list)
        self.__norm_data_is_computed = True

    def normalize_vect(self, x_vect, minus_lb=True):
        """Normalizes a vector of the design space. Unbounded variables are not
        normalized.

        :param x_vect: design variables
        :type x_vect: ndarray
        :param minus_lb: if True, remove lower bounds at normalization
        :return: normalized vector
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        # Normalize the relevant components:
        norm_vect = array(x_vect, copy=True)
        # In case lb=ub
        norm_inds = self.__norm_inds

        if len(x_vect.shape) == 1:
            if minus_lb:
                norm_vect[norm_inds] -= self.__lower_bounds_array[norm_inds]
            norm_vect[norm_inds] /= self._norm_factor[norm_inds]
            # In case lb=ub put value to 0
        elif len(x_vect.shape) == 2:
            if minus_lb:
                norm_vect[:, norm_inds] -= self.__lower_bounds_array[norm_inds]
            norm_vect[:, norm_inds] /= self._norm_factor[norm_inds]
        else:
            raise ValueError("The array to be normalized must be 1d or 2d.")

        to_zero = self.__to_zero
        if to_zero.size > 0:
            norm_vect[to_zero] = 0.0
        return norm_vect

    def unnormalize_vect(self, x_vect, minus_lb=True, no_check=False):
        """Unnormalizes a normalized vector of the design space.

        :param x_vect: design variables
        :type x_vect: ndarray
        :param minus_lb: if True, remove lower bounds at normalization
        :param no_check: if True, don't check that values are in [0,1]
        :return: normalized vector
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        norm_inds = self.__norm_inds
        l_bounds = self.__lower_bounds_array
        if not no_check:
            # Get the indexes of the components to be unnormalized:
            # Check whether the input components are between 0 and 1:
            if (x_vect < -1e-12).any() or (x_vect > 1 + 1e-12).any():
                msg = "All components of the "
                msg += "normalized vector should be between 0 and 1."
                lb_viol = x_vect[x_vect < -1e-12]
                if lb_viol.size != 0:
                    msg += " lower bounds violated : " + str(lb_viol)
                ub_viol = x_vect[x_vect > 1 + 1e-12]
                if ub_viol.size != 0:
                    msg += " upper bounds violated : " + str(ub_viol)

                LOGGER.warning(msg)
        # Unnormalize the relevant components:
        unnorm_vect = array(x_vect, copy=True)

        n_dims = len(x_vect.shape)
        if n_dims == 1:
            unnorm_vect[norm_inds] *= self._norm_factor[norm_inds]
            if minus_lb:
                unnorm_vect[norm_inds] += l_bounds[norm_inds]
        elif n_dims == 2:
            unnorm_vect[:, norm_inds] *= self._norm_factor[norm_inds]
            if minus_lb:
                unnorm_vect[:, norm_inds] += l_bounds[norm_inds]
        else:
            raise ValueError(
                "The array to be unnormalized must be 1d or 2d"
                + ", got "
                + str(n_dims)
                + "d !"
            )
        inds_fixed = self.__to_zero
        if inds_fixed.size > 0:
            if n_dims == 1:
                unnorm_vect[inds_fixed] = l_bounds[inds_fixed]
            else:
                unnorm_vect[:, inds_fixed] = l_bounds[inds_fixed]

        r_xvec = self.round_vect(unnorm_vect)
        return r_xvec

    def round_vect(self, x_vect):
        """Rounds the vector where variables are of integer type.

        :param x_vect: design variables to round
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        int_vars = self.__int_vars_indices
        if not int_vars.any():
            return x_vect
        n_dims = len(x_vect.shape)
        rounded_x_vect = x_vect.copy()
        if n_dims == 1:
            rounded_x_vect[int_vars] = np_round(rounded_x_vect[int_vars])
        elif n_dims == 2:
            rounded_x_vect[:, int_vars] = np_round(rounded_x_vect[:, int_vars])
        else:
            raise ValueError(
                "The array to be unnormalized must be 1d or 2d"
                + ", got "
                + str(n_dims)
                + "d !"
            )

        return rounded_x_vect

    def get_current_x_normalized(self):
        """Returns the current point normalized.

        :returns: the x vector as array normalized by the bounds
        """
        try:
            current_x = self.get_current_x()
        except KeyError as err:
            raise KeyError("Cannot compute normalized current x since " + err.args[0])
        return self.normalize_vect(current_x)

    def get_current_x_dict(self):
        """Get the current point in the design space.

        :returns: the x vector as a dict, keys are the variable names
            values are the variable vales as np array
        """
        return self._current_x

    def set_current_x(self, current_x):
        """Set the current point.

        :param current_x: the current design vector
        """
        if isinstance(current_x, dict):
            self._current_x = current_x
        elif isinstance(current_x, ndarray):
            if current_x.size != self.dimension:
                raise ValueError(
                    "Invalid current_x, dimension mismatch "
                    + str(self.dimension)
                    + " != "
                    + str(current_x.size)
                )
            self._current_x = self.array_to_dict(current_x)
        elif isinstance(current_x, OptimizationResult):
            if current_x.x_opt.size != self.dimension:
                raise ValueError(
                    "Invalid x_opt, dimension mismatch "
                    + str(self.dimension)
                    + " != "
                    + str(current_x.x_opt.size)
                )
            x_array = self.array_to_dict(current_x.x_opt)
            self._current_x = x_array
        else:
            raise TypeError(
                "The current should be an array or a dict. "
                + "Got "
                + str(type(current_x))
                + " instead."
            )
        self._check_current_x()

    def set_current_variable(self, name, current_value):
        """Set the current value of a single variable.

        :param name: name of the variable
        :param current_value: current value of the variable
        """
        if name in self.variables_names:
            self._current_x[name] = current_value
        else:
            raise ValueError("Variable " + str(name) + " is not in the design space !")

    def get_size(self, name):
        """Get the size of a variable Return None if the variable is not known.

        :param name: name of the variable
        """
        return self.variables_sizes.get(name, None)

    def get_type(self, name):
        """Get the type of a variable Return None if the variable is not known.

        :param name: name of the variable
        """
        return self.variables_types.get(name, None)

    def get_lower_bound(self, name):
        """Gets the lower bound of a variable.

        :param name: variable name
        :returns: variable lower bound (possibly infinite)
        """
        return self._lower_bounds.get(name)

    def get_upper_bound(self, name):
        """Gets the upper bound of a variable.

        :param name: variable name
        :returns: variable upper bound (possibly infinite)
        """
        return self._upper_bounds.get(name)

    def get_lower_bounds(self, variables_names=None):
        """Generates an array of the variables' lower bounds.

        :param variables_names: names of the variables of which the lower
            bounds are required
        """
        if self.__norm_data_is_computed and variables_names is None:
            return self.__lower_bounds_array
        return self.dict_to_array(self._lower_bounds, all_var_list=variables_names)

    def get_upper_bounds(self, variables_names=None):
        """Generates an array of the variables' upper bounds.

        :param variables_names: names of the variables of which the upper
            bounds are required
        """
        if self.__norm_data_is_computed and variables_names is None:
            return self.__upper_bounds_array
        return self.dict_to_array(self._upper_bounds, all_var_list=variables_names)

    def set_lower_bound(self, name, lower_bound):
        """Set a new lower bound for variable name.

        :param name: name of the variable
        :param lower_bound: lower bound
        """
        if name not in self.variables_names:
            raise ValueError("Variable " + name + " does not exist")

        self._add_bound(name, self.variables_sizes[name], lower_bound, is_lower=True)
        self._add_norm_policy(name)

    def set_upper_bound(self, name, upper_bound):
        """Set a new upper bound for variable name.

        :param name: name of the variable
        :param upper_bound: upper bound
        """
        if name not in self.variables_names:
            raise ValueError("Variable " + name + " does not exist")

        self._add_bound(name, self.variables_sizes[name], upper_bound, is_lower=False)
        self._add_norm_policy(name)

    def array_to_dict(self, x_array):
        """Split the current point into a dictionary with variables names.

        :param x_array: x array to be converted to a dict of array
        """
        x_dict = {}
        current_index = 0
        # order given by self.variables_names
        for name in self.variables_names:
            size = self.variables_sizes[name]
            x_dict[name] = x_array[current_index : current_index + size]
            current_index += size
        return x_dict

    @staticmethod
    def __get_common_dtype(x_dict):
        """If dict has a value complex array, returns numpy.complex128 if dict has real
        values and mixed floats/int, returns numpy.float64 if dict has only int values,
        returns numpy.int32.

        :param x_dict : dictionary of variables
        """
        has_float = False
        has_int = False
        for val_arr in x_dict.values():
            if not isinstance(val_arr, ndarray):
                raise TypeError("x_dict values must be ndarray")
            if val_arr.dtype == complex128:
                return complex128
            if val_arr.dtype == int32:
                has_int = True
            if val_arr.dtype == float64:
                has_float = True
        if has_float:
            return float64
        if has_int:
            return int32
        return float64

    def dict_to_array(self, x_dict, all_vars=True, all_var_list=None):
        """Aggregate a point as dictionary into array.

        :param x_dict: point as dictionary
        :param all_vars: if True, all variables shall be in x_dict
        :param all_var_list: list of whole set of variables,
            if None, use self.variables_names
        """
        dtype = self.__get_common_dtype(x_dict)
        if all_var_list is None:
            all_var_list = self.variables_names
        if all_vars:
            array_list = [array(x_dict[name], dtype=dtype) for name in all_var_list]
        else:
            array_list = [
                array(x_dict[name], dtype=dtype)
                for name in all_var_list
                if name in x_dict
            ]
        return concatenate(array_list)

    def get_pretty_table(self, fields=None):
        """Builds a PrettyTable object from the design space data.

        :param fields: list of fields to export, by default all
        :returns:  the pretty table object
        """
        if fields is None:
            fields = self.TABLE_NAMES
        table = PrettyTable(fields)
        table.float_format = "%.16g"
        for name in self.variables_names:
            size = self.variables_sizes[name]
            l_b = self._lower_bounds.get(name)
            u_b = self._upper_bounds.get(name)
            var_type = self.variables_types[name]
            curr = self._current_x.get(name)
            for i in range(size):
                data = {
                    "name": name,
                    "value": None,
                    "lower_bound": float("-inf"),
                    "upper_bound": float("inf"),
                    "type": var_type[i],
                }
                if l_b is not None and l_b[i] is not None:
                    data["lower_bound"] = l_b[i]
                if u_b is not None and u_b[i] is not None:
                    data["upper_bound"] = u_b[i]
                if curr is not None:
                    data["value"] = curr[i]
                table.add_row([data[key] for key in fields])

        table.align["name"] = "l"
        table.align["type"] = "l"
        return table

    def export_hdf(self, file_path, append=False):
        """Export to hdf file.

        :param file_path: path to file to write
        :param append: if True, appends the data in the file
        """
        mode = "a" if append else "w"

        with h5py.File(file_path, mode) as h5file:
            design_vars_grp = h5file.require_group(self.DESIGN_SPACE_GROUP)
            design_vars_grp.create_dataset(
                self.NAMES_GROUP, data=array(self.variables_names, dtype=string_)
            )

            for name in self.variables_names:
                var_grp = design_vars_grp.require_group(name)
                var_grp.create_dataset(self.SIZE_GROUP, data=self.variables_sizes[name])

                l_b = self._lower_bounds.get(name)
                if l_b is not None:
                    var_grp.create_dataset(self.LB_GROUP, data=l_b)

                u_b = self._upper_bounds.get(name)
                if u_b is not None:
                    var_grp.create_dataset(self.UB_GROUP, data=u_b)

                var_type = self.variables_types[name]
                if var_type is not None:
                    data_array = string_array(var_type)
                    var_grp.create_dataset(
                        self.VAR_TYPE_GROUP, data=data_array, dtype=data_array.dtype
                    )

                value = self._current_x.get(name)
                if value is not None:
                    var_grp.create_dataset(self.VALUE_GROUP, data=self.__to_real(value))

    def import_hdf(self, file_path):
        """Imports design space from hdf file.

        :param file_path:
        """
        with h5py.File(file_path, "r") as h5file:
            design_vars_grp = get_hdf5_group(h5file, self.DESIGN_SPACE_GROUP)
            variables_names = get_hdf5_group(design_vars_grp, self.NAMES_GROUP)

            for name in variables_names:
                name = name.decode()
                var_group = get_hdf5_group(design_vars_grp, name)
                l_b = self.__read_opt_attr_array(var_group, self.LB_GROUP)
                u_b = self.__read_opt_attr_array(var_group, self.UB_GROUP)
                var_type = self.__read_opt_attr_array(var_group, self.VAR_TYPE_GROUP)
                value = self.__read_opt_attr_array(var_group, self.VALUE_GROUP)
                size = get_hdf5_group(var_group, self.SIZE_GROUP)[()]
                self.add_variable(name, size, var_type, l_b, u_b, value)

        self.check()

    @staticmethod
    def __read_opt_attr_array(var_group, dataset_name):
        """Reads an array in a group, can be optional If data does not exists, returns
        None.

        :param var_group : the variable group
        :param dataset_name : name of the data
        :returns: the data as np array, or None
        """
        inval = var_group.get(dataset_name)
        if inval is not None:
            inval = array(inval)
        return inval

    @staticmethod
    def __to_real(data):
        """Convert complex to real numpy array."""
        return array(array(data, copy=False).real, dtype=float64)

    def to_complex(self):
        """Casts the current value to complex."""
        for name, val in self._current_x.items():
            self._current_x[name] = array(val, dtype=complex128)

    def export_to_txt(self, output_file, fields=None, header_char="", **table_options):
        """Exports the design space to a text file.

        :param output_file: output file path
        :param fields: list of fields to export, by default all
        """
        table = self.get_pretty_table(fields)
        table.border = False
        for option, val in table_options.items():
            table.__setattr__(option, val)
        with open(output_file, "w") as outf:
            table_str = header_char + table.get_string()
            outf.write(table_str)

    @staticmethod
    def read_from_txt(input_file, header=None):
        """Parses a csv file to read the DesignSpace.

        :param input_file: returns: s: the design space
        :param header: fields list, or by default, read in the file
        :returns:  the design space
        """
        float_data = genfromtxt(input_file, dtype="float")
        str_data = genfromtxt(input_file, dtype="str")
        if header is None:
            header = strings_to_unicode_list(str_data[0, :].tolist())
            start_read = 1
        else:
            start_read = 0
        if not set(DesignSpace.MINIMAL_FIELDS).issubset(set(header)):
            raise ValueError(
                "Malformed DesignSpace input file "
                + str(input_file)
                + " does not contain minimal "
                + "variables in header :"
                + str(DesignSpace.MINIMAL_FIELDS)
                + ", got instead : "
                + str(header)
            )
        col_map = {field: i for i, field in enumerate(header)}
        var_names = strings_to_unicode_list(str_data[start_read:, 0].tolist())
        unique_names = []
        prev_name = None
        for name in var_names:  # set([]) does not preserve order !
            if name not in unique_names:
                unique_names.append(name)
                prev_name = name
            elif prev_name != name:
                raise ValueError(
                    "Malformed DesignSpace input file "
                    + str(input_file)
                    + " contains some variables ("
                    + name
                    + ") in a non-consecutive order "
                )

        k = start_read
        design_space = DesignSpace()
        lower_bounds_field = DesignSpace.MINIMAL_FIELDS[1]
        upper_bounds_field = DesignSpace.MINIMAL_FIELDS[2]
        value_field = DesignSpace.TABLE_NAMES[2]
        var_type_field = DesignSpace.TABLE_NAMES[-1]
        for name in unique_names:
            size = var_names.count(name)
            l_b = float_data[k : k + size, col_map[lower_bounds_field]]
            u_b = float_data[k : k + size, col_map[upper_bounds_field]]
            if value_field in col_map:
                value = float_data[k : k + size, col_map[value_field]]
            else:
                value = None
            if var_type_field in col_map:
                var_type = str_data[k : k + size, col_map[var_type_field]].tolist()
            else:
                var_type = None
            design_space.add_variable(name, size, var_type, l_b, u_b, value)
            k += size
        design_space.check()
        return design_space

    def __str__(self):
        return "Design Space:\n" + self.get_pretty_table().get_string()

    def project_into_bounds(self, x_c, normalized=False):
        """Projects x_c onto the bounds, using a simple coordinate wise approach.

        :param normalized: if True then the vector is assumed to be normalized
        :type normalized: bool
        :param x_c: x vector (np array)
        :returns: projected x_c
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        if not normalized:
            l_b = self.__lower_bounds_array
            u_b = self.__upper_bounds_array
        else:
            l_b = zeros_like(x_c)
            u_b = ones_like(x_c)
        x_p = array(x_c)
        l_inds = where(x_c < l_b)
        x_p[l_inds] = l_b[l_inds]

        u_inds = where(x_c > u_b)
        x_p[u_inds] = u_b[u_inds]
        return x_p

    def __contains__(self, variable):
        """Return True if a variable is in the design space.

        :param str variable: variable name.
        """
        return variable in self.variables_names

    def __len__(self):
        """Return the length of the design space which is equal to the number of
        variables."""
        return len(self.variables_names)

    def __getitem__(self, variable):
        """Return the data associated with a given variable: name, type, size, lower
        bound, upper bound and current value.

        :param variable: variable name or index.
        :type variable: str or int
        :return: data associated with variable.
        :rtype: dict
        """
        if isinstance(variable, int):
            try:
                variable = self.variables_names[variable]
            except IndexError:
                raise ValueError(
                    "The parameter indices are comprise between 0 and {}. "
                    "Got {}.".format(len(self) - 1, variable)
                )
        else:
            if variable not in self.variables_names:
                raise ValueError(
                    "The design space does not contain '{}'.".format(variable)
                )
        try:
            value = self.get_current_x([variable])
        except KeyError:
            value = None
        return {
            self.NAME_GROUP: variable,
            self.TYPE_GROUP: self.get_type(variable)[0],
            self.VALUE_GROUP: value,
            self.SIZE_GROUP: self.get_size(variable),
            self.LB_GROUP: self.get_lower_bound(variable),
            self.UB_GROUP: self.get_upper_bound(variable),
        }

    def extend(self, other):
        """Extend the design space with another design space.

        :param other: design space to be appended
        :type other: DesignSpace
        """
        for name in other.variables_names:
            size = other.get_size(name)
            var_type = other.get_type(name)
            l_b = other.get_lower_bound(name)
            u_b = other.get_upper_bound(name)
            value = other.get_current_x_dict()[name]
            self.add_variable(name, size, var_type, l_b, u_b, value)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Generate a N2 and XDSM into files (and/or web page) from an Excel description of the
MDO problem."""
from __future__ import division, unicode_literals

import logging
from ast import literal_eval
from typing import Iterable, List, Mapping, Optional, Tuple

from pandas import DataFrame  # noqa F401
from pandas import read_excel

from gemseo.api import (
    create_design_space,
    create_scenario,
    generate_n2_plot,
    get_available_formulations,
)
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.discipline import MDODiscipline
from gemseo.core.mdo_scenario import MDOScenario

LOGGER = logging.getLogger(__name__)


class XLSStudyParser(object):
    """Parse the input Excel file that describe the GEMSEO study.

    The Excel file must contain one sheet per discipline.
    The name of the sheet shall have the name of the discipline.
    The sheet shall have at least two columns,
    one for the inputs and one for the outputs,
    with the following format:

    +--------+---------+----------------+--------------------+----------------+
    | Inputs | Outputs |Design variables| Objective function |  Constraints   |
    +========+=========+================+====================+================+
    |  in1   |  out1   |      in1       |       out1         |     out2       |
    +--------+---------+----------------+--------------------+----------------+
    |  in2   |  out2   |                |                    |                |
    +--------+---------+----------------+--------------------+----------------+

    Empty lines are ignored.
    All the objective functions and constraints must be outputs of a discipline,
    not necessarily the one of the current sheet.
    All the design variables must be inputs of a discipline,
    not necessarily the one of the current sheet.

    Attributes:
        xls_study_path (str): The path to the Excel file.
        frames (Dict[str,DataFrame]): The data frames created from the Excel file.
        disciplines (Dict[str,MDODiscipline]): The disciplines.
        scenarios (Dict[str,Dict[str,Union[str,List[str]]]]): The descriptions
            of the scenarios parsed in the Excel file.
        inputs (Set[str]): The input variables.
        outputs (Set[str]): The output variables.
    """

    SCENARIO_PREFIX = "Scenario"
    DISCIPLINE = "Discipline"
    DISCIPLINES = "Disciplines"
    OBJECTIVE_FUNCTION = "Objective function"
    CONSTRAINTS = "Constraints"
    DESIGN_VARIABLES = "Design variables"
    FORMULATION = "Formulation"
    OPTIONS = "Options"
    OPTIONS_VALUES = "Options values"

    def __init__(
        self, xls_study_path  # type: str
    ):  # type (..) -> None
        """Initialize the study from the Excel specification.

        Args:
            xls_study_path: The path to the Excel file describing the study.

        Raises:
            IOError: If the Excel file cannot be opened.
            ValueError: If no scenario has been found in Excel file.
        """
        self.xls_study_path = xls_study_path
        try:
            self.frames = read_excel(xls_study_path, sheet_name=None, engine="openpyxl")
        except IOError:
            LOGGER.error("Failed to open the study file: %s", xls_study_path)
            raise

        LOGGER.info("Detected the following disciplines: %s", list(self.frames.keys()))
        self.disciplines = dict()
        self.scenarios = dict()
        self.inputs = dict()
        self.outputs = dict()

        self._init_disciplines()
        self._get_opt_pb_descr()

        if not self.scenarios:
            raise ValueError("No scenario found in the xls file")

    def _init_disciplines(self):  # type: (...) -> None
        """Initialize the disciplines.

        Raises:
            ValueError: If the discipline has no input column or output column.
        """
        all_inputs = []
        all_outputs = []
        for disc_name, frame in self.frames.items():
            if disc_name.startswith(self.SCENARIO_PREFIX):
                continue
            LOGGER.info("Parsing discipline %s", disc_name)

            missing_column_msg = (
                "The sheet of the discipline '{}' must have a column '{}'"
            )

            try:
                inputs = self._get_frame_series_values(frame, "Inputs")
            except ValueError:
                raise ValueError(missing_column_msg.format(disc_name, "Inputs"))

            all_inputs += inputs
            try:
                outputs = self._get_frame_series_values(frame, "Outputs")
            except ValueError:
                raise ValueError(missing_column_msg.format(disc_name, "Outputs"))

            all_outputs += outputs
            disc = MDODiscipline(disc_name)
            disc.input_grammar.initialize_from_data_names(inputs)
            disc.output_grammar.initialize_from_data_names(outputs)
            LOGGER.info("Inputs: %s", inputs)
            LOGGER.info("Outputs: %s", outputs)

            self.disciplines[disc_name] = disc

        self.inputs = set(all_inputs)
        self.outputs = set(all_outputs)

    @staticmethod
    def _get_frame_series_values(
        frame,  # type: DataFrame
        series_name,  # type: str
        return_none=False,  # type: Optional[bool]
    ):  # type: (...) -> None
        """Return the data of a named column.

        Removes empty data.

        Args:
            frame: The pandas frame of the sheet.
            series_name: The name of the series.
            return_none: If the series does not exists, returns None
                instead of raising a ValueError.

        Returns:
            The list of a named column.

        Raises:
            ValueError: If the sheet has no name.
        """
        series = frame.get(series_name)
        if series is None:
            if return_none:
                return None
            raise ValueError("The sheet has no serie named '{}'".format(series_name))
        # Remove empty data
        # pylint: disable=comparison-with-itself
        return [val for val in series.tolist() if val == val]

    def _get_opt_pb_descr(self):  # type: (...) -> None
        """Initialize the objective function, constraints and design_variables.

        Raises:
            ValueError: If at least one of following elements is missing:
                * ``disciplines`` column,
                * ``design variables`` column,
                * ``objectives`` column,
                * ``constraints`` column,
                * ``formulations`` column,
                * if a scenario has more than one formulation,
                * if a scenario has different number of option values.
        """
        self.scenarios = dict()

        for frame_name, frame in self.frames.items():
            if not frame_name.startswith(self.SCENARIO_PREFIX):
                continue
            LOGGER.info("Detected scenario in sheet: %s", frame_name)

            missing_column_msg = "Scenario {} has no {} column!"

            try:
                disciplines = self._get_frame_series_values(frame, self.DISCIPLINES)
            except ValueError:
                raise ValueError(
                    missing_column_msg.format(frame_name, self.DISCIPLINES)
                )

            try:
                design_variables = self._get_frame_series_values(
                    frame, self.DESIGN_VARIABLES
                )
            except ValueError:
                raise ValueError(
                    missing_column_msg.format(frame_name, self.DESIGN_VARIABLES)
                )

            try:
                objectives = self._get_frame_series_values(
                    frame, self.OBJECTIVE_FUNCTION
                )
            except ValueError:
                raise ValueError(
                    missing_column_msg.format(frame_name, self.OBJECTIVE_FUNCTION)
                )

            try:
                constraints = self._get_frame_series_values(frame, self.CONSTRAINTS)
            except ValueError:
                raise ValueError(
                    missing_column_msg.format(frame_name, self.CONSTRAINTS)
                )

            try:
                formulation = self._get_frame_series_values(frame, self.FORMULATION)
            except ValueError:
                raise ValueError(
                    missing_column_msg.format(frame_name, self.FORMULATION)
                )

            options = self._get_frame_series_values(frame, self.OPTIONS, True)
            options_values = self._get_frame_series_values(
                frame, self.OPTIONS_VALUES, True
            )

            if len(formulation) != 1:
                raise ValueError(
                    "Scenario {} must have one {} value!".format(
                        str(frame_name), self.FORMULATION
                    )
                )

            if options is not None:
                if len(options) != len(options_values):
                    raise ValueError(
                        "Options {} and Options values {} "
                        "must have the same length!".format(options, options_values)
                    )

            formulation = formulation[0]

            scn = {
                self.DISCIPLINES: disciplines,
                self.OBJECTIVE_FUNCTION: objectives,
                self.CONSTRAINTS: constraints,
                self.DESIGN_VARIABLES: design_variables,
                self.FORMULATION: formulation,
                self.OPTIONS: options,
                self.OPTIONS_VALUES: options_values,
            }

            self.scenarios[frame_name] = scn

        for name, desc in self.scenarios.items():
            self._check_opt_pb(
                desc[self.OBJECTIVE_FUNCTION],
                desc[self.CONSTRAINTS],
                desc[self.DISCIPLINES],
                desc[self.DESIGN_VARIABLES],
                desc[self.FORMULATION],
                name,
            )

    def _check_opt_pb(
        self,
        objectives,  # type: Iterable[str]
        constraints,  # type: Iterable[str]
        disciplines,  # type: Iterable[str]
        design_variables,  # type: Iterable[str]
        formulation,  # type: str
        scn_name,  # type: str
    ):  # type: (...) -> None
        """Checks the optimization problem consistency.

        Args:
            objectives: The names of the objectives.
            constraints: The names of the constraints.
            disciplines: The names of the disciplines.
            design_variables: The names of the design variables.
            formulation: The name of the MDO formulation.
            scn_name: The name of the scenario.

        Raises:
            ValueError: If at least one of following situation happens:
                * design variables in the scenario are not input of any discipline,
                * some disciplines do not exist in the scenario,
                * some constraints are not outputs of any discipline,
                * the objective function is not an output of any discipline,
                * the formulation is unknown.
        """
        LOGGER.info("New scenario: %s", scn_name)
        LOGGER.info("Objectives: %s", objectives)
        LOGGER.info("Disciplines: %s", disciplines)
        LOGGER.info("Constraints: %s", constraints)
        LOGGER.info("Design variables: %s", design_variables)
        LOGGER.info("Formulation: %s", formulation)

        missing = set(design_variables) - self.inputs
        if missing:
            raise ValueError(
                "{}: some design variables are not "
                "the inputs of any discipline: {}".format(scn_name, list(missing))
            )

        missing = (
            set(disciplines)
            - set(list(self.disciplines.keys()))
            - set(list(self.scenarios))
        )
        if missing:
            raise ValueError(
                "{}: some disciplines don't exist: {}".format(
                    scn_name, str(list(missing))
                )
            )

        missing = set(constraints) - self.outputs
        if missing:
            raise ValueError(
                "Some constraints of {} are not outputs of any discipline: {}".format(
                    scn_name, list(missing)
                )
            )

        missing = set(objectives) - self.outputs
        if missing:
            raise ValueError(
                "Some objectives of {} are not "
                "outputs of any discipline: {}".format(scn_name, list(missing))
            )
        if not objectives:
            raise ValueError("No objectives of {} are defined".format(scn_name))

        if formulation not in get_available_formulations():
            raise ValueError(
                "Unknown formulation '{}'; use one of: {}".format(
                    formulation, get_available_formulations()
                )
            )


class StudyAnalysis(object):
    """A MDO study analysis from an Excel specification.

    Generate a N2 (equivalent to the Design Structure Matrix) diagram,
    showing the couplings between the disciplines,
    and a XDSM (Extended Design Structure Matrix),
    showing the MDO process,
    from an Excel specification of the inputs, outputs, design variables,
    objectives and constraints.

    The input Excel files contains one sheet per discipline.
    The name of the sheet shall have the discipline name.
    The sheet shall have at least two columns,
    one for inputs and one for outputs,
    with the following format:

    .. table:: Disc1

        +--------+---------+
        | Inputs | Outputs |
        +========+=========+
        |  in1   |  out1   |
        +--------+---------+
        |  in2   |  out2   |
        +--------+---------+

    Empty lines are ignored.

    The scenarios (at least one, or multiple for distributed formulations)
    must appear in a Excel sheet name starting by "Scenario".

    The sheet shall have the following columns,
    with some constraints :

    - All of them are mandatory, even if empty for the constraints.
    - The order does not matter.
    - One and only one formulation must be declared.
    - At least one objective must be provided, and one design variable.

    .. table:: Scenario1

        +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+
        |Design variables| Objective function |  Constraints   |  Disciplines   |  Formulation   |  Options       | Options values |
        +================+====================+================+================+================+================+================+
        |      in1       |       out1         |     out2       |     Disc1      |     MDF        |  tolerance     |     0.1        |
        +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+
        |                |                    |                |     Disc2      |                | main_mda_class |   MDAJacobi    |
        +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+

    All the objective functions and constraints must be outputs of a discipline,
    not necessarily the one of the current sheet.
    All the design variables must be inputs of a discipline,
    not necessarily the one of the current sheet.

    The columns 'Options' and 'Options values' are used to pass the formulation options.
    Note that for string type 'Option values',
    the value can be written with or without the "" characters.

    To use multi level MDO formulations,
    create multiple scenarios,
    and add the name of the sub scenarios
    in the list of disciplines of the main (system) scenario.

    An arbitrary number of levels can be generated this way
    (three, four levels etc formulations).

    Attributes:
        xls_study_path (str): The path of the Excel file.
        study (XLSStudyParser): The XLSStudyParser instance built from the Excel file.
        disciplines_descr (Dict[str,MDODiscipline]): The descriptions of the disciplines
            (including sub-scenario) parsed in the Excel file.
        scenarios_descr (Dict[str,Dict[str,Union[str,List[str]]]]): The descriptions
            of the scenarios parsed in the Excel file.
        disciplines (Dict[str,MDODiscipline]): The disciplines.
        scenarios (Dict[str,MDOScenario]): The scenarios.
    """  # noqa: B950

    AVAILABLE_DISTRIBUTED_FORMULATIONS = ("BiLevel", "BLISS98B")

    def __init__(
        self, xls_study_path  # type: str
    ):  # type: (...) -> None
        """Initialize the study from the Excel specification.

        Args:
            xls_study_path: The path to the Excel file describing the study.
        """
        self.xls_study_path = xls_study_path
        self.study = XLSStudyParser(self.xls_study_path)
        self.disciplines_descr = self.study.disciplines
        self.scenarios_descr = self.study.scenarios
        self.disciplines = dict()
        self.scenarios = dict()
        self.main_scenario = None
        self._create_scenarios()

    def generate_n2(
        self,
        file_path="n2.pdf",  # type: str
        show_data_names=True,  # type: bool
        save=True,  # type: bool
        show=False,  # type:bool
        figsize=(15, 10),  # type: Tuple[float, float]
    ):  # type: (...) -> None
        """Generate a N2 plot for the disciplines list.

        Args:
            file_path: The file path of the figure.
            show_data_names: If true,
                the names of the coupling data is shown;
                otherwise,
                circles are drawn,
                which size depends on the number of coupling names.
            save: If True, save the figure to file_path.
            show: If True, show the plot.
            figsize: The size of the figure.
        """
        generate_n2_plot(
            list(self.disciplines.values()),
            file_path,
            show_data_names,
            save,
            show,
            figsize,
        )

    @staticmethod
    def _create_scenario(
        disciplines,  # type: Iterable[MDODiscipline]
        scenario_descr,  # type: Mapping[str, Iterable[str]]
    ):  # type: (...) -> MDOScenario
        """Create a MDO scenario.

        Args:
            disciplines: The discipline.
            scenario_descr: The description of the scenario.

        Returns:
            A MDO scenario.
        """
        coupl_struct = MDOCouplingStructure(disciplines)
        couplings = coupl_struct.get_all_couplings()
        design_space = create_design_space()
        scn_dv = scenario_descr[XLSStudyParser.DESIGN_VARIABLES]
        for var in sorted(set(scn_dv) | set(couplings)):
            design_space.add_variable(var, size=1)

        options = scenario_descr[XLSStudyParser.OPTIONS]
        options_dict = {}
        if options is not None:
            options_values = scenario_descr[XLSStudyParser.OPTIONS_VALUES]
            for opt, val in zip(options, options_values):
                if isinstance(val, str):
                    try:
                        val = literal_eval(val)
                    except ValueError:
                        pass
                else:
                    pass
                options_dict[opt] = val

        scenario = create_scenario(
            disciplines,
            scenario_descr[XLSStudyParser.FORMULATION],
            scenario_descr[XLSStudyParser.OBJECTIVE_FUNCTION],
            design_space,
            **options_dict
        )
        for cstr in scenario_descr[XLSStudyParser.CONSTRAINTS]:
            scenario.add_constraint(cstr)
        return scenario

    def _get_disciplines_instances(
        self, scn  # type: Mapping[str, Iterable[str]]
    ):  # type: (...) -> List[MDODiscipline]
        """Get the instances of the disciplines from a scenario.

        Args:
            scn: The description of the scenario.

        Returns:
            The instances of the disciplines of the scenario.
        """
        discs = []
        for disc_name in scn[XLSStudyParser.DISCIPLINES]:
            disc_inst = self.disciplines_descr.get(disc_name)
            if disc_inst is None:  # not a discipline, so maybe a scenario
                disc_inst = self.scenarios.get(disc_name)
                if disc_inst is None:
                    return None
            discs.append(disc_inst)
        return discs

    def _create_scenarios(self):  # type: (...) -> None
        """Create the main scenario, eventually including sub scenarios.

        Raises:
            ValueError: If crossed dependencies exist between scenarios.
        """
        n_scn = len(self.scenarios_descr)
        i = 0

        temp_discs = {}
        while len(self.scenarios) != n_scn and i <= n_scn:
            i += 1
            for name, scn in self.scenarios_descr.items():
                discs = self._get_disciplines_instances(scn)
                if discs is not None:  # All depdendencies resolved
                    for disc in discs:
                        if not disc.is_scenario():
                            temp_discs[disc.name] = disc

                    scenario = self._create_scenario(discs, scn)
                    self.scenarios[name] = scenario
                    # The last scenario created is the one
                    # with the most dependencies
                    # so the main one
                    self.main_scenario = scenario
        # At each while iteration at least 1 scenario must be resolved
        # otherwise this means there is a cross dependency between
        # scenarios
        if len(self.scenarios) != n_scn:
            raise ValueError(
                "Scenarios dependencies cannot be resolved, "
                "check for cycling dependencies between scenarios!"
            )

        self.disciplines = dict()
        # Preserves the order of disciplines in the original Excel file
        # Important for the N2 generation
        for disc_name in self.disciplines_descr:
            self.disciplines[disc_name] = temp_discs[disc_name]

    def generate_xdsm(
        self,
        output_dir,  # type: str
        latex_output=False,  # type: bool
        open_browser=False,  # type: bool
    ):  # type: (...) -> MDOScenario
        """Create an xdsm.json file from the current scenario.

        Args:
            output_dir: The directory where the XDSM html files are generated.
            latex_output: If True, build the .tex, .tikz and .pdf files.
            open_browser: If True, open in a web browser.

        Returns:
            The MDOScenario that contains the DesignSpace, the
            formulation, but the disciplines have only correct
            input and output grammars but no _run methods so that can't be executed
        """
        LOGGER.info("Generated the following Scenario:")
        LOGGER.info("%s", self.main_scenario)
        LOGGER.info("%s", self.main_scenario.formulation.opt_problem)
        self.main_scenario.xdsmize(
            outdir=output_dir, latex_output=latex_output, open_browser=open_browser
        )
        return self.main_scenario



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Benoit Pauwels - Stacked data management
#               (e.g. iteration index)

"""
A database of function calls and design variables
*************************************************
"""

from __future__ import division, unicode_literals

import logging
from ast import literal_eval
from hashlib import sha1
from itertools import chain, islice
from xml.etree.ElementTree import parse as parse_element

import h5py
from numpy import array, atleast_2d, concatenate, float64, ndarray, string_, uint8
from numpy.linalg import norm
from six import string_types

from gemseo.utils.ggobi_export import save_data_arrays_to_xml
from gemseo.utils.hdf5 import get_hdf5_group
from gemseo.utils.py23_compat import OrderedDict

LOGGER = logging.getLogger(__name__)


class Database(object):
    """Class to store evaluations of functions, such as DOE or optimization histories.

    Avoids multiple calls of the same functions,
    useful when simulations are costly

    It is also used to store inputs and retrieve them
    for optimization graphical post processing and plots
    generation

    Can be serialized to HDF5 for portability and cold post processing
    """

    missing_value_tag = "NA"
    KEYSSEPARATOR = "__KEYSSEPARATOR__"
    GRAD_TAG = "@"
    ITER_TAG = "Iter"

    def __init__(self, input_hdf_file=None):
        """Constructor."""
        self.__dict = OrderedDict()
        self.__max_iteration = 0
        # Call functions when store is called

        if input_hdf_file is not None:
            self.import_hdf(input_hdf_file)

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """Sets an item of the dictionary.

        :param key: the key of the item
        :param value: the value of the item
        :param dict_setitem: the set item default method
        """
        if not isinstance(key, (ndarray, HashableNdarray)):
            raise TypeError(
                "Optimization history keys must be design variables numpy arrays"
            )
        if not isinstance(value, dict):
            raise TypeError("Optimization history values must be data dictionary")
        if isinstance(key, HashableNdarray):
            self.__dict[key] = value
        else:
            self.__dict[HashableNdarray(key, True)] = value

    @staticmethod
    def __get_hashed_key(x_vect):
        """Gets the HashableNdarray from x_vect.

        :param x_vect: the x design vector or a HashableNdarray
        :returns : the HashableNdarray of x_vect
        """
        if not isinstance(x_vect, (ndarray, HashableNdarray)):
            raise TypeError(
                "Optimization history keys must be design variables numpy arrays"
            )
        if isinstance(x_vect, ndarray):
            return HashableNdarray(x_vect)
        return x_vect

    def __getitem__(self, x_vect):
        hashed = self.__get_hashed_key(x_vect)
        return self.__dict[hashed]

    def __delitem__(self, x_vect):
        hashed = self.__get_hashed_key(x_vect)
        del self.__dict[hashed]

    def setdefault(self, key, default):
        """Sets a default database entry."""
        if not isinstance(key, (ndarray, HashableNdarray)):
            raise TypeError(
                "Optimization history keys must be design variables numpy arrays"
            )
        if not isinstance(default, dict):
            raise TypeError("Optimization history values must be data dictionary")
        return self.__dict.setdefault(key, default)

    def __len__(self):
        return len(self.__dict)

    def keys(self):
        """Database keys generator."""
        for key in self.__dict.keys():
            yield key

    def values(self):
        """Database values generator."""
        for value in self.__dict.values():
            yield value

    def items(self):
        """Database items generator."""
        for key, val in self.__dict.items():
            yield key, val

    def get_value(self, x_vect):
        """Accessor for the values.

        :param x_vect: the design variables
        """
        return self[x_vect]

    def get_max_iteration(self):
        """Maximum iteration number."""
        return self.__max_iteration

    def get_x_history(self):
        """Get list of x ordered by calls.

        :returns: the list of x np arrays
        """
        return [x_vect.unwrap() for x_vect in self.__dict.keys()]

    def get_last_n_x(self, n):
        """Get list of n last x ordered by calls.

        :returns: the list of x np arrays
        """
        n_max = len(self)
        if n > n_max:
            raise ValueError(
                "n ={} must be lower than the database size = {}".format(n, n_max)
            )
        return [
            x_vect.unwrap() for x_vect in islice(self.__dict.keys(), n_max - n, n_max)
        ]

    def get_index_of(self, x_vect):
        """Returns the index of a particular x.

        :param x_vect: x numpy array
        :returns: the index of x_vect, or throws a key error
        """
        hashed = HashableNdarray(x_vect)
        for i, key in enumerate(self.__dict.keys()):
            if key == hashed:
                return i
        raise KeyError(x_vect)

    def get_x_by_iter(self, iteration):
        """Return design variables at a specified iteration.

        :param iteration: the iteration number
        :returns: the numpy array of x at iteration
        """
        nkeys = len(self.__dict)
        if nkeys == 0:
            raise ValueError("The database is empty!")
        if iteration < 0:
            iteration = nkeys + iteration
        if iteration >= nkeys or (iteration < 0 and -iteration > nkeys):
            raise ValueError(
                "iteration should be lower than"
                + " maximum iteration = "
                + str(len(self) - 1)
                + " got instead : "
                + str(iteration)
            )
        for i, key in enumerate(self.__dict.keys()):
            if i == iteration:
                return key.unwrap()

    def clear(self):
        """Clears the database."""
        self.__dict.clear()

    def clean_from_iterate(self, iterate):
        """Delete the iterates after a given iterate number.

        :param iterate: the iterate number
        """

        def gen_todel():
            for iterate_number, x_vect in enumerate(self.__dict.keys()):
                # remove iterations beyond limit iterate number
                if iterate < iterate_number:
                    yield x_vect

        # Copies only the keys after iterate
        to_del = list(gen_todel())
        for key in to_del:
            del self.__dict[key]
        self.__max_iteration = len(self)

    def remove_empty_entries(self):
        """Removes empty entries, when x is associated to an empty dict."""
        empt = [
            k
            for k, v in self.items()
            if len(v) == 0 or (len(v) == 1 and list(v.keys())[0] == self.ITER_TAG)
        ]
        for k in empt:
            del self[k]

    def filter(self, data_list_to_keep):
        """Keeps only the values in the data list.

        :param data_list_to_keep: the list of data names to keep
        """
        data_list_to_keep = set(data_list_to_keep)
        for val in self.values():
            keys_to_del = set(val.keys()) - data_list_to_keep
            for key in keys_to_del:
                del val[key]

    def get_func_history(self, funcname, x_hist=False):
        """Return function values history. Can also return history of design variables.

        :param funcname: the function name
        :param x_hist: if True, returns variables history as well
            (Default value = False)
        :returns: the function history list
        """
        outf_l = []
        x_history = []
        for x_vect, out_val in self.items():
            val = out_val.get(funcname)
            if val is not None:
                if isinstance(val, ndarray) and val.size == 1:
                    val = val[0]
                outf_l.append(val)
                if x_hist:
                    x_history.append(x_vect.unwrap())
        outf = array(outf_l)
        if x_hist:
            return outf, x_history

        return outf

    def get_func_grad_history(self, funcname, x_hist=False):
        """Return gradient values history Can also return history of design variables.

        :param funcname: the function name
        :param x_hist: if True, returns variables history as well
            (Default value = False)
        :returns: the jacobian history list
        """
        gradient_name = self.get_gradient_name(funcname)
        return self.get_func_history(funcname=gradient_name, x_hist=x_hist)

    def is_func_grad_history_empty(self, funcname):
        """Check if history is empty.

        :param funcname: the function name
        :returns: True if history is empty
        """
        return len(self.get_func_grad_history(funcname, x_hist=False)) == 0

    def contains_x(self, x_vect):
        """Tests if history has a design variables x stored.

        :param x_vect: the design variables to test
        :returns: True if x_vect is in self
        """
        return HashableNdarray(x_vect) in self.__dict

    def get_f_of_x(self, fname, x_vect, dist_tol=0.0):
        """If x in self, get associated "fname" value, if it exists.

        :param fname: the function name
        :param x_vect: the design variables
        :returns: the values associated to x with name fname
        """
        if dist_tol == 0.0:
            vals = self.get(x_vect)
            if vals is not None:
                return vals.get(fname)  # Returns None if not in self
        else:
            for x_key, vals in self.items():
                x_v = x_key.unwrap()
                if norm(x_v - x_vect) <= dist_tol * norm(x_v):
                    return vals.get(fname)
        return None

    def get(self, x_vect, default=None):
        """Return the value for key if key is in the dictionary, else default."""
        if not isinstance(x_vect, (HashableNdarray, ndarray)):
            raise TypeError(
                "Optimization history keys must be design variables numpy arrays"
            )
        if isinstance(x_vect, ndarray):
            x_vect = HashableNdarray(x_vect)
        return self.__dict.get(x_vect, default)

    def pop(self, k):
        """D.pop(k[,d]) -> v, remove specified key and return the corresponding value.

        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        return self.__dict.pop(k)

    def contains_dataname(self, data_name, skip_grad=False):
        """Tests if history has a value named data_name stored.

        :param data_name: the name of the data
        :param skip_grad: do not account for gradient names
        :returns: True if data_name is in self
        """
        return data_name in self.get_all_data_names(skip_grad=skip_grad)

    def store(self, x_vect, values_dict, add_iter=True):
        """Stores the values associated to the variables x.

        :param x_vect: design variables vector
        :param values_dict: values to be stored
        :param add_iter: add iteration information
            (Default value = True)
        """
        if self.contains_x(x_vect):
            curr_val = self.get_value(x_vect)
            # No new keys = already computed = new iteration
            # otherwise just calls to other functions
            curr_val.update(values_dict)
        elif add_iter:
            self.__max_iteration += 1
            # include the iteration index
            new_values_dict = dict(
                values_dict, **{self.ITER_TAG: [self.__max_iteration]}
            )
            self.__setitem__(x_vect, new_values_dict)
        else:
            self.__max_iteration += 1
            # do not include the iteration index but still update it
            self.__setitem__(x_vect, values_dict)

    def get_all_data_names(self, skip_grad=True, skip_iter=False):
        """Return data variables (design, functions, gradient, ... Gradient variables
        can be skipped.

        :param skip_grad: do not list gradient names (Default value = True)
        :param skip_iter: do not add Iter in the list

        :returns: the list of data names in the database
        """
        names = set()
        for value in self.__dict.values():
            for key in value.keys():
                if skip_grad and key.startswith(self.GRAD_TAG):
                    continue
                names.add(key)
        if skip_iter and self.ITER_TAG in names:
            names.remove(self.ITER_TAG)
        return sorted(names)

    def _format_history_names(self, functions, stacked_data):
        """Formats the functions names to be displayed in the history.

        :param functions: param stacked_data:
        :param stacked_data:
        """
        if functions is None:
            functions = self.get_all_data_names()
        if stacked_data is None:
            if self.ITER_TAG in functions:
                stacked_data = [self.ITER_TAG]
            else:
                stacked_data = iter([])
        elif not set(stacked_data).issubset(functions):
            raise ValueError(
                "The names of the data to be unstacked ("
                + str(stacked_data)
                + ")"
                + " must be included in the names of the data"
                + " to be returned ("
                + str(functions)
                + ")."
            )
        elif self.ITER_TAG in functions and self.ITER_TAG not in stacked_data:
            stacked_data.append(self.ITER_TAG)
        return functions, stacked_data

    def get_complete_history(
        self,
        functions=None,
        add_missing_tag=False,
        missing_tag="NA",
        all_iterations=False,
        stacked_data=None,
    ):
        """Return complete history of optimization: design variables, functions,
        gradients.

        :param functions: functions names to get (Default value = None)
        :param add_missing_tag: add "missing_tag" when data is not available
            for this iteration (Default value = False)
        :param missing_tag: the missing tag to add (Default value = 'NA')
        :param all_iterations: if True, points called at several
            iterations will be duplicated in the history (each duplicate
            corresponding to a different calling index); otherwise each point
            will appear only once (with the latest calling index)
            (Default value = False)
        :param stacked_data: list of names corresponding to data stored as
            lists. For example the iterations indexes are stored in a list.
            Other examples of
            stacked data may be penalization parameters or trust region radii.
            (Default value = None)
        :returns: function history and x history as lists
        """
        functions, stacked_data = self._format_history_names(functions, stacked_data)
        f_history = []
        x_history = []
        for x_vect, out_val in self.items():
            # If duplicates are not to be considered, or if no iteration index
            # is specified, then only one entry (the last one) will be written:
            if not all_iterations or self.ITER_TAG not in out_val:
                first_index = -1
                last_index = -1
            # Otherwise all the entries will be written:
            else:
                first_index = 0
                # N.B. if the list of indexes is empty, then no entry will be
                # written.
                last_index = len(out_val[self.ITER_TAG]) - 1
            # Add an element to the history for each duplicate required:
            for duplicate_ind in range(first_index, last_index + 1):
                out_vals = []
                for funcname in functions:
                    if funcname in out_val:
                        if funcname not in stacked_data:
                            out_vals.append(out_val[funcname])
                        # If the data 'funcname' is stacked and there remains
                        # entries to unstack, then unstack the next entry:
                        elif duplicate_ind < len(out_val[funcname]):
                            val = out_val[funcname][duplicate_ind]
                            out_vals.append(val)
                    elif add_missing_tag:
                        out_vals.append(missing_tag)
                if out_vals:
                    f_history.append(out_vals)
                    x_history.append(x_vect.unwrap())
        return f_history, x_history

    @staticmethod
    def __to_real(data):
        """Convert complex to real numpy array."""
        return array(array(data, copy=False).real, dtype=float64)

    def export_hdf(self, file_path="optimization_history.h5", append=False):
        """Export optimization history to hdf file.

        :param file_path: path to file to write
            (Default value = 'optimization_history.h5')
        :param append: if True, appends the data in the file
            (Default value = False)
        """
        mode = "a" if append else "w"

        with h5py.File(file_path, mode) as h5file:
            design_vars_grp = h5file.require_group("x")
            keys_group = h5file.require_group("k")
            values_group = h5file.require_group("v")
            iterated = self.items()
            i = 0

            if append and design_vars_grp:
                iterated = islice(iterated, len(design_vars_grp), len(self.__dict))
                i = len(design_vars_grp)

            for key, val in iterated:
                design_vars_grp.create_dataset(str(i), data=key.unwrap())
                keys_data = array(list(val.keys()), dtype=string_)
                locvalues_scalars = []
                argrp = None

                for ind, locval in enumerate(val.values()):
                    if isinstance(locval, (ndarray, list)):
                        if argrp is None:
                            argrp = values_group.require_group("arr_" + str(i))
                        argrp.create_dataset(str(ind), data=self.__to_real(locval))
                    else:
                        locvalues_scalars.append(locval)

                keys_group.create_dataset(str(i), data=keys_data)
                values_group.create_dataset(
                    str(i), data=self.__to_real(locvalues_scalars)
                )
                i += 1

    def import_hdf(self, filename="optimization_history.h5"):
        """Imports a database from hdf file.

        :param filename: Default value = 'optimization_history.h5')
        """
        with h5py.File(filename, "r") as h5file:
            design_vars_grp = h5file["x"]
            keys_group = h5file["k"]
            values_group = h5file["v"]

            for raw_index in range(len(design_vars_grp)):
                str_index = str(raw_index)
                keys = [k.decode() for k in get_hdf5_group(keys_group, str_index)]

                array_name = "arr_{}".format(str_index)

                if array_name in values_group:
                    argrp = values_group[array_name]
                    vec_dict = {keys[int(k)]: array(v) for k, v in argrp.items()}
                else:
                    vec_dict = {}

                locvalues_scalars = get_hdf5_group(values_group, str_index)
                scalar_keys = (k for k in keys if k not in vec_dict)

                scalar_dict = dict(
                    ((k, v) for k, v in zip(scalar_keys, locvalues_scalars))
                )
                scalar_dict.update(vec_dict)

                self.store(
                    array(design_vars_grp[str_index]), scalar_dict, add_iter=False
                )

    @staticmethod
    def set_dv_names(n_dv):
        """Create a list of default design variables names.

        :param n_dv: number of design variables in problem
        :returns: a list of design variables names
        """
        return ["x_" + str(i) for i in range(1, n_dv + 1)]

    def _format_design_variables_names(self, design_variables_names, dimension):
        """Formats the design variables names to be displayed in the history.

        :param design_variables_names: param dimension:
        :param dimension: number of components
        """
        if design_variables_names is None:
            design_variables_names = self.set_dv_names(dimension)
        elif isinstance(design_variables_names, string_types):
            design_variables_names = [design_variables_names]
        elif not isinstance(design_variables_names, list) and not isinstance(
            design_variables_names, tuple
        ):
            raise TypeError(
                "design_variables_names must be a list or a tuple: "
                "a {} is provided".format(type(design_variables_names))
            )
        return design_variables_names

    def get_history_array(
        self,
        functions=None,
        design_variables_names=None,
        add_missing_tag=False,
        missing_tag="NA",
        add_dv=True,
        all_iterations=False,
        stacked_data=None,
    ):
        """Return history of optimization process.

        :param functions: functions names to export (Default value = None)
        :param design_variables_names: names of the design variables
            (Default value = None)
        :param missing_tag: missing tag to add (Default value = 'NA')
        :param add_dv: if True, adds the design variables to the
            returned array (Default value = True)
        :param add_missing_tag: add "missing_tag" when data is not available
            for this iteration (Default value = False)
        :param missing_tag: the missing tag to add (Default value = 'NA')
        :param all_iterations: if True, points called at several
            iterations will be duplicated in the history (each duplicate
            corresponding to a different calling index); otherwise each point
            will appear only once (with the latest calling index)
            (Default value = False)
        :param stacked_data: list of names corresponding to data stored as
            lists. For example the iterations indexes are stored in a list.
            Other examples of
            stacked data may be penalization parameters or trust region radii.
            (Default value = None)
        :returns: function history and x history as lists
        """
        if functions is None:
            functions = self.get_all_data_names()
        f_history, x_history = self.get_complete_history(
            functions, add_missing_tag, missing_tag, all_iterations, stacked_data
        )
        design_variables_names = self._format_design_variables_names(
            design_variables_names, len(x_history[0])
        )
        flat_vals = []
        fdict = OrderedDict()
        for f_val_i in f_history:
            flat_vals_i = []
            for f_val, f_name in zip(f_val_i, functions):
                if isinstance(f_val, list):
                    f_val = array(f_val)
                if isinstance(f_val, ndarray) and len(f_val) > 1:
                    flat_vals_i = flat_vals_i + f_val.tolist()
                    fdict[f_name] = [
                        f_name + "_" + str(i + 1) for i in range(len(f_val))
                    ]
                else:
                    flat_vals_i.append(f_val)
                    if f_name not in fdict:
                        fdict[f_name] = [f_name]
            flat_vals.append(flat_vals_i)
        flat_names = sorted(list(chain(*fdict.values())))

        x_flat_vals = []
        xdict = OrderedDict()
        for x_val_i in x_history:
            x_flat_vals_i = []
            for x_val, x_name in zip(x_val_i, design_variables_names):
                if isinstance(x_val, ndarray) and len(x_val) > 1:
                    x_flat_vals_i = x_flat_vals_i + x_val.tolist()
                    xdict[x_name] = [
                        x_name + "_" + str(i + 1) for i in range(len(x_val))
                    ]
                else:
                    x_flat_vals_i.append(x_val)
                    if x_name not in xdict:
                        xdict[x_name] = [x_name]
            x_flat_vals.append(x_flat_vals_i)

        x_flat_names = list(chain(*xdict.values()))
        if add_dv:
            variables_names = flat_names + x_flat_names
        else:
            variables_names = flat_names

        f_history = array(flat_vals).real
        x_history = array(x_flat_vals).real
        if add_dv:
            f2d = atleast_2d(f_history)
            x2d = atleast_2d(x_history)
            if f2d.shape[0] == 1:
                f2d = f2d.T
            if x2d.shape[0] == 1:
                x2d = x2d.T
            values_array = concatenate((f2d, x2d), axis=1)
        else:
            values_array = f_history
        return values_array, variables_names, functions

    def export_to_ggobi(
        self, functions=None, file_path="opt_hist.xml", design_variables_names=None
    ):
        """Export history to xml file format for ggobi tool.

        :param functions: Default value = None)
        :param file_path: Default value = "opt_hist.xml")
        :param design_variables_names: Default value = None)
        """
        values_array, variables_names, functions = self.get_history_array(
            functions, design_variables_names, add_missing_tag=True, missing_tag="NA"
        )
        LOGGER.info("Export to ggobi for functions: %s", str(functions))
        LOGGER.info("Export to ggobi file: %s", str(file_path))
        save_data_arrays_to_xml(
            variables_names=variables_names,
            values_array=values_array,
            file_path=file_path,
        )

    def import_from_opendace(self, database_file):
        """Reads an opendace xml database.

        :param database_file: the path to the database file
        """
        tree = parse_element(database_file)
        for link in tree.getroot().iter("link"):
            data = {}
            for information in link:
                for x_ydyddy in information:
                    data[x_ydyddy.tag] = literal_eval(x_ydyddy.text)
            x_vect = array(data.pop("x"))
            data_reformat = data["y"]
            for key, value in data["dy"].items():
                data_reformat["@" + key[1:]] = array(value)
            self.store(x_vect, data_reformat)

    @classmethod
    def get_gradient_name(
        cls,
        name,  # type: str
    ):  # type: (...) -> str
        """Return the name of the gradient related to a function.

        This name is the concatenation of a GRAD_TAG, e.g. '@',
        and the name of the function, e.g. 'f'.
        With this example, the name of the gradient is '@f'.

        Args:
            name: The name of a function.

        Returns:
            The name of the gradient based on the name of the function.
        """
        return "{}{}".format(cls.GRAD_TAG, name)


class HashableNdarray(object):

    """HashableNdarray wrapper for ndarray objects.

    Instances of ndarray are not HashableNdarray,
    meaning they cannot be added to
    sets, nor used as keys in dictionaries. This is by design - ndarray
    objects are mutable, and therefore cannot reliably implement the
    __hash__() method.

    The HashableNdarray class allows a way around this limitation.
    It implements the required methods for HashableNdarray
    objects in terms of an encapsulated
    ndarray object. This can be either a copied instance (which is safer)
    or the original object (which requires the user to be careful enough
    not to modify it).
    """

    def __init__(self, wrapped, tight=False):
        """Creates a new HashableNdarray object encapsulating an ndarray.

        :param wrapped:The wrapped ndarray.
        :param tight: If True, a copy of the input ndaray is created.
        """
        self.__tight = tight
        self.wrapped = array(wrapped) if tight else wrapped
        self.__hash = int(sha1(wrapped.view(uint8)).hexdigest(), 16)

    def __eq__(self, other):
        return all(self.wrapped == other.wrapped)

    def __hash__(self):
        return self.__hash

    def unwrap(self):
        """Returns the encapsulated ndarray.

        If the wrapper is "tight", a copy of the encapsulated ndarray is
        """
        if self.__tight:
            return array(self.wrapped)

        return self.wrapped

    def __str__(self):
        return str(array(self.wrapped))

    def __repr__(self):
        return str(self)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable problem - Problem
**************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import zeros
from numpy.random import rand
from numpy.random import seed as npseed

from .design_space import TMDesignSpace
from .models import TMMainModel, TMSubModel
from .variables import (
    X_SHARED_NAME,
    check_consistency,
    get_constraint_name,
    get_coupling_name,
    get_u_local_name,
    get_x_local_name,
)

LOGGER = logging.getLogger(__name__)


class TMProblem(object):

    """The scalable problem from Tedford and Martins, 2010, builds a list of strongly
    coupled models (:class:`.TMSubModel`) completed by a main model
    (:class:`.TMMainModel`) computing the objective function and the constraints.

    These disciplines are defined on an unit design space whose parameters comprised in
    [0, 1] (:class:`.TMDesignSpace`). This problem is defined by the number of shared
    design parameters, the number of local design parameters per discipline and the
    number of output coupling variables per discipline. The strongly coupled disciplines
    can be either fully coupled (one discipline depends on all the others) or circularly
    coupled (one discipline depends only on the previous one and the first discipline
    depends only on the last one).
    """

    def __init__(
        self,
        n_shared=1,
        n_local=None,
        n_coupling=None,
        full_coupling=True,
        noised_coupling=False,
        seed=1,
    ):
        """Constructor.

        :param int n_shared: size of the shared design parameters.
            Default: 1.
        :param list(int) n_local: sizes of the local design parameters for the different
            disciplines. Same length as n_coupling. If None, use [1, 1]. Default: None.
        :param list(int) n_coupling: sizes of the coupling parameters for the different
            disciplines. Same length as n_local. If None, use [1, 1]. Default: None.
        :param bool full_coupling: if True, fully couple the disciplines. Otherwise,
            use circular coupling. Default: True.
        :param bool noised_coupling: if True, add a noise component u_local_i
            on the i-th discipline output.
        :param int seed: seed for replicability.
        """
        npseed(seed)
        self._seed = seed

        # Set the coupling style
        self._full_coupling = full_coupling
        self._noised_coupling = noised_coupling

        # Set and check the dimensions of the problem
        n_local = n_local or [1, 1]
        n_coupling = n_coupling or [1, 1]
        self.n_shared = n_shared
        self.n_local = n_local
        self.n_coupling = n_coupling
        check_consistency(n_shared, n_local, n_coupling)
        self.n_submodels = len(n_local)

        # Create instances of the random coefficients
        c_shared, c_local, c_cpl, c_constraint = self._generate_coefficients()

        # Instantiate the system model
        names = [X_SHARED_NAME]
        names += [get_coupling_name(index) for index in range(self.n_submodels)]
        default_inputs = self.get_default_inputs(names=names)
        self.models = [self._create_main_model(c_constraint, default_inputs)]

        # Instantiate the strongly coupled models
        for index in range(self.n_submodels):
            names = [X_SHARED_NAME]
            names += [get_x_local_name(index)]
            if full_coupling:
                names += [
                    get_coupling_name(other_index)
                    for other_index in range(self.n_submodels)
                    if other_index != index
                ]
            else:
                other_id = self.n_submodels - 1 if index == 0 else index - 1
                names += [get_coupling_name(other_id)]
            if self._noised_coupling:
                names += [get_u_local_name(index)]
            default_inputs = self.get_default_inputs(names=names)
            model = self._create_sub_model(
                index, c_shared[index], c_local[index], c_cpl[index], default_inputs
            )
            self.models.append(model)

        # Instantiate the design space
        self.design_space = self.get_design_space()

    @classmethod
    def _create_main_model(cls, c_constraint, default_inputs):
        """Create main model.

        :param ndarray c_constraint: coefficients for constraint.
        :param dict(ndarray) default_inputs: default inputs.
        :return: instance of the main model.
        :return: TMMainModel
        """
        return TMMainModel(c_constraint, default_inputs)

    @classmethod
    def _create_sub_model(cls, index, c_shared, c_local, c_cpl, default_inputs):
        """Create sub model.

        :param int index: model index
        :param ndarray c_shared: coefficients for shared design parameters.
        :param ndarray c_local: coefficients for local design parameters.
        :param ndarray c_cpl: coefficients for coupling variables.
        :param dict(ndarray) default_inputs: default inputs.
        :return: instance of a sub-model.
        :return: TMSubModel
        """
        return TMSubModel(index, c_shared, c_local, c_cpl, default_inputs)

    def __str__(self):
        """String representation."""
        msg = ["Scalable problem"]
        for model in self.models:
            msg.append(".... {}".format(model.name))
            msg.append("........ Inputs:")
            for name in model.inputs_names:
                size = model.inputs_sizes[name]
                msg.append("............ {} ({})".format(name, size))
            msg.append("........ Outputs:")
            for name in model.outputs_names:
                size = model.outputs_sizes[name]
                msg.append("............ {} ({})".format(name, size))
        return "\n".join(msg)

    def get_default_inputs(self, names=None):
        """Get default input values.

        :param list(str) names: names of the inputs.
        :return: name and values of the inputs.
        :rtype: dict
        """
        inputs = {X_SHARED_NAME: zeros(self.n_shared) + 0.5}
        for index in range(self.n_submodels):
            inputs[get_x_local_name(index)] = zeros(self.n_local[index]) + 0.5
            inputs[get_coupling_name(index)] = zeros(self.n_coupling[index]) + 0.5
            inputs[get_constraint_name(index)] = zeros(self.n_coupling[index]) + 0.5
            inputs[get_u_local_name(index)] = zeros(self.n_coupling[index])
        if names is not None:
            inputs = {name: inputs[name] for name in names}
        return inputs

    def get_design_space(self):
        """Get the TM design space.

        :return: instance of the design space
        :rtype: TMDesignSpace
        """
        return TMDesignSpace(self.n_shared, self.n_local, self.n_coupling)

    def reset_design_space(self):
        """Reset the TM design space."""
        self.design_space = self.get_design_space()

    def _generate_coefficients(self):
        """Generate coefficients associated with the shared design parameters, the local
        design parameters and the coupling variables.

        :return: coefficients for both shared design parameters,
            local design parameters and coupling variables.
        :rtype: list(ndarray), list(ndarray), list(ndarray), list(ndarray)
        """
        c_shared = []
        c_local = []
        c_coupling = []
        for disc in range(self.n_submodels):
            if self._full_coupling:
                other_indices = set(range(self.n_submodels)) - {disc}
                other_indices = list(other_indices)
            else:
                other_index = self.n_submodels - 1 if disc == 0 else disc - 1
                other_indices = [other_index]
            n_coupling = self.n_coupling[disc]
            c_shared.append(rand(n_coupling, self.n_shared))
            c_local.append(rand(n_coupling, self.n_local[disc]))
            c_coupling.append({})
            for index in other_indices:
                coeff = rand(n_coupling, self.n_coupling[index])
                c_coupling[-1][get_coupling_name(index)] = coeff
        c_constraint = [rand(val) for val in self.n_coupling]
        return c_shared, c_local, c_coupling, c_constraint



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import division, unicode_literals

import inspect
import logging
from os import remove
from os.path import abspath, basename
from os.path import dirname as pdirname
from os.path import exists
from typing import Mapping, Union

from six import string_types

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequenceFactory
from gemseo.core.function import MDOFunction
from gemseo.formulations.formulations_factory import MDOFormulationsFactory
from gemseo.post.post_factory import PostFactory
from gemseo.utils.string_tools import MultiLineString, pretty_repr

"""
Base class for all Scenarios
****************************
"""


LOGGER = logging.getLogger(__name__)

ScenarioInputDataType = Mapping[str, Union[str, int, Mapping[str, Union[int, float]]]]


class Scenario(MDODiscipline):
    """Base class for MDO and DOE scenarios.

    Multidisciplinary Design Optimization Scenario, main user interface
    Creates an optimization problem and solves it with a driver

    MDO Problem description: links the disciplines and the formulation
    to create an optimization problem.
    Use the class by instantiation.
    Create your disciplines beforehand.

    Specify the formulation by giving the class name such as the string
    "MDF"

    The reference_input_data is the typical input data dict that is provided
    to the run method of the disciplines

    Specify the objective function name, which must be an output
    of a discipline of the scenario, with the "objective_name" attribute

    If you want to add additional design constraints,
    use the add_constraint method

    To view the results, use the "post_process" method after execution.
    You can view:

    - The design variables history, the objective value, the constraints,
      by using:
      scenario.post_process("OptHistoryView", show=False, save=True)
    - Quadratic approximations of the functions close to the
      optimum, when using gradient based algorithms, by using:
      scenario.post_process("QuadApprox", method="SR1", show=False,
      save=True, function="my_objective_name",
      file_path="appl_dir")
    - Self Organizing Maps of the design space, by using:
      scenario.post_process("SOM", save=True, file_path="appl_dir")

    To list post processings on your setup,
    use the method scenario.posts
    For more detains on their options, go to the "gemseo.post" package
    """

    # Constants for input variables in json schema
    X_0 = "x_0"
    U_BOUNDS = "u_bounds"
    L_BOUNDS = "l_bounds"
    ALGO = "algo"
    ALGO_OPTIONS = "algo_options"

    def __init__(
        self,
        disciplines,
        formulation,
        objective_name,
        design_space,
        name=None,
        **formulation_options
    ):
        """Constructor, initializes the MDO scenario Objects instantiation and checks
        are made before run intentionally.

        :param disciplines: the disciplines of the scenario
        :param formulation: the formulation name,
            the class name of the formulation in gemseo.formulations
        :param objective_name: the objective function name
        :param design_space: the design space
        :param name: scenario name
        :param formulation_options: options for creation of the formulation
        """
        self.formulation = None
        self.formulation_name = None
        self.disciplines = disciplines
        self.optimization_result = None
        self._algo_factory = None

        self._check_disciplines()
        self._init_algo_factory()
        self._form_factory = self._formulation_factory
        super(Scenario, self).__init__(name=name)
        self._init_base_grammar(self.__class__.__name__)
        self._init_formulation(
            formulation, objective_name, design_space, **formulation_options
        )
        self.post_factory = PostFactory()
        self._update_input_grammar()

    @property
    def _formulation_factory(self):
        """Returns formulations factory."""
        return MDOFormulationsFactory()

    def _check_disciplines(self):
        """Check that two disciplines dont compute the same output."""

        all_outs = set()
        for disc in self.disciplines:
            outs = set(disc.get_output_data_names())
            common = outs & all_outs
            if len(common) > 0:
                msg = "Two disciplines, among which {}, compute the same output: {}"
                raise ValueError(msg.format(disc.name, common))
            all_outs = all_outs | outs

    @property
    def design_space(self):
        """Proxy for formulation.design_space.

        :returns: the design space
        """
        return self.formulation.design_space

    def _init_base_grammar(self, name):
        """Initializes the base grammars from MDO scenario inputs and outputs This
        ensures that subclasses have base scenario inputs and outputs Can be overloaded
        by subclasses if this is not desired.

        :param name: name of the scenario, used as base name for the json
            schema to import: name_input.json and name_output.json
        """
        comp_dir = abspath(pdirname(inspect.getfile(Scenario)))
        input_grammar_file = self.auto_get_grammar_file(True, name, comp_dir)
        output_grammar_file = self.auto_get_grammar_file(False, name, comp_dir)
        self._instantiate_grammars(input_grammar_file, output_grammar_file)

    def set_differentiation_method(self, method="user", step=1e-6):
        """Sets the differentiation method for the process.

        :param method: the method to use, either "user", "finite_differences",
            or "complex_step" or "no_derivatives",
            which is equivalent to None. (Default value = "user")
        :param step: Default value = 1e-6)
        """
        if method is None:
            method = "no_derivatives"
        self.formulation.opt_problem.differentiation_method = method
        self.formulation.opt_problem.fd_step = step

    def add_constraint(
        self,
        output_name,
        constraint_type=MDOFunction.TYPE_EQ,
        constraint_name=None,
        value=None,
        positive=False,
        **kwargs
    ):
        """Add a user constraint, i.e. a design constraint in addition to formulation
        specific constraints such as targets in IDF. The strategy of repartition of
        constraints is defined in the formulation class.

        :param output_name: the output name to be used as constraint
            for instance, if g_1 is given and
            constraint_type="eq",
            g_1=0 will be added as constraint to the optimizer
            If a list is given, a single discipline
            must provide all
            outputs
        :param constraint_type: the type of constraint, "eq" for equality,
            "ineq" for inequality constraint
            (Default value = MDOFunction.TYPE_EQ)
        :param constraint_name: name of the constraint to be stored,
            if None, generated from the output name
            (Default value = None)
        :param value: Default value = None)
        :param positive: Default value = False)
        :returns: the constraint function as an MDOFunction
        """
        if constraint_type not in [MDOFunction.TYPE_EQ, MDOFunction.TYPE_INEQ]:
            raise ValueError(
                "Constraint type must be either 'eq' or 'ineq',"
                + " got:"
                + str(constraint_type)
                + " instead"
            )

        self.formulation.add_constraint(
            output_name,
            constraint_type,
            constraint_name=constraint_name,
            value=value,
            positive=positive,
            **kwargs
        )

    def add_observable(self, output_names, observable_name=None, discipline=None):
        """Add observable to the optimization problem. The repartition strategy of the
        observable is defined in the formulation class. When more than one output name
        is provided, the observable function returns a concatenated array of the output
        values.

        :param output_names: names of the outputs to observe
        :param observable_name: name of the observable, optional. If None, the
            output name is used by default.
        :type observable_name: str
        :param discipline: if None, detected from inner disciplines, otherwise
            the discipline used to build the function
            (Default value = None)
        :type discipline: MDODiscipline
        """
        return self.formulation.add_observable(
            output_names, observable_name, discipline
        )

    def _init_formulation(
        self, formulation, objective_name, design_space, **formulation_options
    ):
        """Initializes the formulation given disciplines, objective name and design
        variables names.

        :param formulation: the formulation name to use
        :param design_space: the design space object
        :param objective_name: the objective function name
        :param formulation_options: options for creation of the formulation
        """
        if not isinstance(formulation, string_types):
            raise TypeError(
                "Formulation must be specified by its name!"
                + " Please use GEMSEO_PATH to specify custom formulations"
            )
        form_inst = self._form_factory.create(
            formulation,
            disciplines=self.disciplines,
            objective_name=objective_name,
            design_space=design_space,
            **formulation_options
        )
        self.formulation_name = formulation
        self.formulation = form_inst

    def get_optim_variables_names(self):
        """A convenience function to access formulation design variables names.

        :returns: the decision variables of the scenario
        :rtype: list(str)
        """
        return self.formulation.get_optim_variables_names()

    def get_optimum(self):
        """Return the optimization results.

        :returns: Optimal solution found by the scenario if executed, None
                  otherwise
        :rtype: OptimizationResult
        """

        return self.optimization_result

    def save_optimization_history(
        self, file_path, file_format=OptimizationProblem.HDF5_FORMAT, append=False
    ):
        """Saves the optimization history of the scenario to a file.

        :param file_path: The path to the file to save the history
        :param file_format: The format of the file, either "hdf5" or "ggobi"
            (Default value = "hdf5")
        :param append: if True, data is appended to the file if not empty
            (Default value = False)
        """
        opt_pb = self.formulation.opt_problem
        if file_format == OptimizationProblem.HDF5_FORMAT:
            opt_pb.export_hdf(file_path=file_path, append=append)
        elif file_format == OptimizationProblem.GGOBI_FORMAT:
            opt_pb.database.export_to_ggobi(file_path=file_path)
        else:
            raise ValueError(
                "Cannot export optimization history"
                + " to file format:"
                + str(file_format)
            )

    def set_optimization_history_backup(
        self,
        file_path,
        each_new_iter=False,
        each_store=True,
        erase=False,
        pre_load=False,
        generate_opt_plot=False,
    ):
        """Sets the backup file for the optimization history during the run.

        :param file_path: The path to the file to save the history
        :param each_new_iter: if True, callback at every iteration
        :param each_store: if True, callback at every call to store()
            in the database
        :param erase: if True, the backup file is erased before the run
        :param pre_load: if True, the backup file is loaded before run,
            useful after a crash
        :param generate_opt_plot: generates the optimization history view
            at backup
        """
        opt_pb = self.formulation.opt_problem

        if exists(file_path):
            if erase and pre_load:
                raise ValueError(
                    "Conflicting options for history backup, "
                    + "cannot pre load optimization history"
                    + " and erase it!"
                )
            if erase:
                LOGGER.warning("Erasing optimization history in %s", str(file_path))
                remove(file_path)
            elif pre_load:
                opt_pb.database.import_hdf(file_path)

        def backup_callback():
            """A callback function to backup optimization history."""
            self.save_optimization_history(file_path, append=True)
            if generate_opt_plot and opt_pb.database:
                basepath = basename(file_path).split(".")[0]
                self.post_process(
                    "OptHistoryView", save=True, show=False, file_path=basepath
                )

        opt_pb.add_callback(
            backup_callback, each_new_iter=each_new_iter, each_store=each_store
        )

    @property
    def posts(self):
        """Lists the available post processings.

        :returns: the list of methods
        """
        return self.post_factory.posts

    def post_process(self, post_name, **options):
        """Finds the appropriate library and executes the post processing on the
        problem.

        :param post_name: the post processing name
        :param options: options for the post method, see its package
        """
        post = self.post_factory.execute(
            self.formulation.opt_problem, post_name, **options
        )
        return post

    def _run_algorithm(self):
        """Runs the algo, either DOE or optimizer."""
        raise NotImplementedError()

    def __repr__(self):
        msg = MultiLineString()
        msg.add(self.name)
        msg.indent()
        msg.add("Disciplines: {}", pretty_repr(self.disciplines, delimiter=" "))
        msg.add("MDOFormulation: {}", self.formulation.__class__.__name__)
        msg.add("Algorithm: {}", self.local_data.get(self.ALGO))
        return str(msg)

    def get_disciplines_statuses(self):
        """Retrieves the disciplines statuses.

        :returns: the statuses dict, key: discipline name, value: status
        """
        statuses = {}
        for disc in self.disciplines:
            statuses[disc.__class__.__name__] = disc.status
        return statuses

    def print_execution_metrics(self):
        """Prints total number of executions and cumulated runtime by discipline."""
        n_lin = 0
        n_calls = 0
        LOGGER.info("* Scenario Executions statistics *")
        for disc in self.disciplines:
            LOGGER.info("* Discipline: %s", disc.name)
            LOGGER.info("Executions number: %s", str(disc.n_calls))
            LOGGER.info("Execution time:  %s s", str(disc.exec_time))

            n_calls += disc.n_calls
            LOGGER.info("Linearizations number: %s", str(disc.n_calls_linearize))

            n_lin += disc.n_calls_linearize
        LOGGER.info("Total number of executions calls %s", str(n_calls))
        LOGGER.info("Total number of linearizations %s", str(n_lin))

    def xdsmize(
        self,
        monitor=False,
        outdir=".",
        print_statuses=False,
        outfilename="xdsm.html",
        latex_output=False,
        open_browser=False,
        html_output=True,
        json_output=False,
    ):
        """Creates an xdsm.json file from the current scenario. If monitor is set to
        True, the xdsm.json file is updated to reflect discipline status update (hence
        monitor name).

        :param bool monitor: if True, updates the generated file at each
            discipline status change
        :param str outdir: the directory where XDSM json file is generated
        :param bool print_statuses: print the statuses in the console at
            each update
        :param outfilename: file name of the output. THe basename
            is used and the extension adapted for the HTML / JSON / PDF
            outputs
        :param bool latex_output: build .tex, .tikz and .pdf file
        :param open_browser: if True, opens the web browser with the XDSM
        :param html_output: if True, outputs a self contained HTML file
        :param json_output: if True, outputs a JSON file for XDSMjs
        """
        from gemseo.utils.xdsmizer import XDSMizer

        if print_statuses:
            monitor = True

        if monitor:
            XDSMizer(self).monitor(outdir=outdir, print_statuses=print_statuses)
        else:
            XDSMizer(self).run(
                output_directory_path=outdir,
                latex_output=latex_output,
                open_browser=open_browser,
                html_output=html_output,
                json_output=json_output,
                outfilename=outfilename,
            )

    def get_expected_dataflow(self):
        """Overriden method from MDODiscipline base class delegated to formulation
        object."""
        return self.formulation.get_expected_dataflow()

    def get_expected_workflow(self):
        """Overriden method from MDODiscipline base class delegated to formulation
        object."""
        exp_wf = self.formulation.get_expected_workflow()
        return ExecutionSequenceFactory.loop(self, exp_wf)

    def _init_algo_factory(self):
        """Initalizes the algorithms factory."""
        raise NotImplementedError()

    def get_available_driver_names(self):
        """Returns the list of available drivers."""
        return self._algo_factory.algorithms

    def _update_input_grammar(self):
        """Updates input grammar from algos names."""

        available_algos = self.get_available_driver_names()
        algo_grammar = {"type": "string", "enum": available_algos}
        self.input_grammar.set_item_value("algo", algo_grammar)

    @staticmethod
    def is_scenario():
        """Retuns True if self is a scenario.

        :returns: True if self is a scenario
        """
        return True



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""A Grammar based on JSON schema."""

from __future__ import division, unicode_literals

import json
import logging
from numbers import Number

from numpy import generic, ndarray, zeros

from gemseo.core.grammars.abstract_grammar import AbstractGrammar
from gemseo.core.grammars.errors import InvalidDataException
from gemseo.core.grammars.json_schema import MutableMappingSchemaBuilder
from gemseo.core.grammars.simple_grammar import SimpleGrammar
from gemseo.utils.py23_compat import PY2, JsonSchemaException, Path, compile_schema

LOGGER = logging.getLogger(__name__)


class JSONGrammar(AbstractGrammar):
    """A grammar subclass that stores the input or output data types and structure a
    MDODiscipline using a JSON format It is able to check the inputs and outputs against
    a JSON schema."""

    PROPERTIES_FIELD = "properties"
    REQUIRED_FIELD = "required"
    TYPE_FIELD = "type"
    OBJECT_FIELD = "object"
    TYPES_MAP = {
        "array": ndarray,
        "float": float,
        "string": str,
        "integer": int,
        "boolean": bool,
        "number": Number,
    }

    def __init__(
        self,
        name,
        schema_file=None,
        schema=None,
        grammar_type=None,
        descriptions=None,
    ):
        """Constructor.

        :param name : grammar name
        :param schema : a genson Schema  to initialize self
        :param schema_file : the json schema file
        """
        super(JSONGrammar, self).__init__(name)
        self._validator = None
        self.schema = None
        self._init_schema()

        if schema is not None:
            self.schema.add_schema(schema)
        elif schema_file is not None:
            self.init_from_schema_file(schema_file, descriptions=descriptions)
        else:
            self.initialize_from_base_dict({}, description_dict=descriptions)

    def __repr__(self):
        return "{}, schema: {}".format(self, self.schema.to_json())

    def _init_schema(self):
        """Initialize the schema."""
        self.schema = MutableMappingSchemaBuilder()

    def clear(self):
        """Clears the data to produce an empty grammar."""
        self.__set_grammar_from_dict({})

    def _init_validator(self):
        """Initialize the validator."""
        schema_dict = self.schema.to_schema()
        schema_dict.pop("id", None)
        self._validator = compile_schema(schema_dict)

    @classmethod
    def cast_array_to_list(cls, data_dict):
        """Casts the numpy arrays to lists for dictionary values.

        :param data_dict: the data dictionary
        :returns: The dict with casted arrays
        """
        out_d = data_dict.copy()
        for key, value in data_dict.items():
            if isinstance(value, (ndarray, generic)):
                out_d[key] = value.real.tolist()
            elif isinstance(value, dict):
                out_d[key] = cls.cast_array_to_list(value)
        return out_d

    def load_data(self, data, raise_exception=True):
        """Loads the data dictionary in the grammar and checks it against json schema.

        :param data: the input data
        :param raise_exception: if False, no exception is raised
            when data is invalid (Default value = True)
        """
        if not isinstance(data, dict):
            raise InvalidDataException(
                "Input data must be a dictionary: got a {} instead".format(type(data))
            )

        if self._validator is None:
            self._init_validator()

        error_exist = False
        data_to_check = self.cast_array_to_list(data)

        try:
            self._validator(data_to_check)
        except JsonSchemaException as error:
            error_exist = True
            msg = "Invalid data in: {}".format(self.name)

            if error.args[0].startswith("data must contain"):
                # Error messages are not clear enough when missing properties
                # All keys are put in the message
                diff = sorted(set(self.get_data_names()) - set(data.keys()))
                if diff:
                    msg += "\nMissing mandatory properties: {}".format(",".join(diff))
                else:
                    msg += "\n', error: {}".format(error.args[0])
            else:
                msg += "\n', error: {}".format(error.args[0])

            LOGGER.error(msg)

        if error_exist and raise_exception:
            err = "Invalid data from grammar {}".format(self.name)
            raise InvalidDataException(err)

        # Check a copy to keep types and arrays but store initial dict for complex
        # Add defaults
        for key, value in data_to_check.items():
            data.setdefault(key, value)

        return data

    def init_from_schema_file(self, schema_path, descriptions=None):
        """Set the grammar from a file.

        :param schema_path: path to the schema file.
        """
        schema_path = Path(schema_path)

        if not schema_path.exists():
            msg = "Try to initialize grammar with not existing file: {}".format(
                schema_path
            )
            raise FileNotFoundError(msg)

        schema_dict = json.loads(schema_path.read_text())
        self.__set_grammar_from_dict(schema_dict, descriptions)

    def __set_grammar_from_dict(self, schema_dict, descriptions=None):
        """Set the grammar from a dictionary.

        Args:
            schema_dict: Schema dictionary.
            descriptions: Properties descriptions, optional.
        """
        self._init_schema()
        self.__update_grammar_from_dict(schema_dict, descriptions)

    def __update_grammar_from_dict(self, schema_dict, descriptions=None):
        """Update the grammar from a dictionary.

        Args:
            schema_dict: Schema dictionary.
            descriptions: Properties descriptions, optional.
        """
        if descriptions is not None:
            if not isinstance(schema_dict, dict):
                schema_dict = schema_dict.to_schema()
            for ppty_name, ppty_schema in schema_dict["properties"].items():
                try:
                    ppty_schema["description"] = descriptions[ppty_name]
                except KeyError:
                    LOGGER.debug(
                        "skipping description for unknown property %s", ppty_name
                    )

        self.__merge_schema(schema_dict)

    def __merge_schema(self, schema):
        """Merge a schema in the current one.

        Args:
            schema: Schema to merge, could be a schema object or a dictionary.
        """
        self.schema.add_schema(schema)
        self._validator = None

    def initialize_from_data_names(
        self,
        data_names,
        descriptions=None,
    ):
        """Initializes a JSONGrammar from a list of data. All data of the grammar will
        be set as arrays.

        :param data_names: a data names list
        """
        data = zeros(1)
        typical_data_dict = {k: data for k in data_names}
        self.initialize_from_base_dict(typical_data_dict, description_dict=descriptions)

    def initialize_from_base_dict(
        self,
        typical_data_dict,
        description_dict=None,
    ):
        """Initialize a json grammar with types and names from a typical data entry. The
        keys of the typical_data_dict will be the names of the data in the grammar. The
        types of the values of the typical_data_dict will be converted to JSON Schema
        types and define the properties of the JSON Schema.

        :param typical_data_dict: a data dictionary with keys as data names
            and values as a typical value for this data
        :param description_dict: dictionary of descriptions,
             {name:meaning} structure
        """
        # Convert arrays to list as for check
        list_data_dict = self.cast_array_to_list(typical_data_dict)
        self.schema.add_object(list_data_dict)
        self.__set_grammar_from_dict(self.schema, description_dict)

    def get_data_names(self):
        """Returns the list of data names.

        :returns: the data names, as a dict keys set
        """
        return list(self.schema.keys())

    def is_data_name_existing(self, data_name):
        """Checks if data_name is present in grammar.

        :param data_name: the data name
        :returns: True if data is in grammar
        """
        return data_name in self.get_data_names()

    def is_all_data_names_existing(self, data_names):
        """Checks if data_names are present in grammar.

        :param data_names: the data names list
        :returns: True if all data are in grammar
        """
        exists = self.is_data_name_existing
        for data_name in data_names:
            if not exists(data_name):
                return False
        return True

    def update_from(self, input_grammar):
        """Adds properties coming from another grammar.

        :param input_grammar: the grammar to take inputs from
        """
        if not isinstance(input_grammar, JSONGrammar):
            msg = (
                "A {} is expected as input, but an object of type {} "
                "has been provided.".format(self.__class__, type(input_grammar))
            )
            raise TypeError(msg)

        self.__merge_schema(input_grammar.schema)

    def to_simple_grammar(self):
        """Creates a SimpleGrammar from self, preserving the features if possible
        Ignores the features of JSONGrammar that are not supported by SimpleGrammar.

        :returns: a SimpleGrammar instance
        """
        grammar = SimpleGrammar(self.name)
        schema_dict = self.schema.to_schema()
        properties = schema_dict.get(self.PROPERTIES_FIELD, {})

        for prop, desc in properties.items():
            grammar.data_names.append(prop)
            if "type" not in desc or desc["type"] not in self.TYPES_MAP:
                d_type = None
            else:
                d_type = self.TYPES_MAP[desc["type"]]
            grammar.data_types.append(d_type)

            if d_type == "array":
                sub_type = desc["items"].get("type")
                if sub_type not in ["number", "integer", None]:
                    msg = "Unsupported type {sub_type} in JSONGrammar {name}"
                    msg += "for property {prop} in conversion to simple grammar"
                    err = msg.format(sub_type=sub_type, name=self.name, prop=prop)
                    LOGGER.warning(err)

            for prop_name in ["minItems", "maxItems", "additionalItems", "contains"]:
                if prop_name in desc:
                    msg = "Unsupported feature {desc} in JSONGrammar {name}"
                    msg += "for property {prop} in conversion to simple grammar"
                    err = msg.format(desc=desc, name=self.name, prop=prop)
                    LOGGER.warning(err)

        return grammar

    def update_from_if_not_in(self, input_grammar, exclude_grammar):
        """Adds objects coming from input_grammar if they are not in exclude_grammar.

        :param input_grammar: the grammar to take inputs from
        :param exclude_grammar: exclusion grammar
        """
        if not (
            isinstance(input_grammar, self.__class__)
            and isinstance(exclude_grammar, self.__class__)
        ):
            msg = self._get_update_error_msg(self, input_grammar, exclude_grammar)
            raise TypeError(msg)

        schema = MutableMappingSchemaBuilder()
        schema.add_schema(input_grammar.schema)

        for name in exclude_grammar.schema.keys():
            try:
                del schema[name]
            except KeyError:
                pass

        self.__merge_schema(schema)

    def restrict_to(self, data_names):
        """Restrict the grammar to given names.

        :param data_names: the names of the data to restrict the grammar to
        """
        for name in list(self.schema.keys()):
            if name not in data_names:
                self.remove_item(name)

    def remove_item(self, item_name):
        """Removes an item from the grammar.

        :param item_name: the item name to be removed
        """
        del self.schema[item_name]

    def set_item_value(self, item_name, item_value):
        """Sets the value of an item.

        :param item_name: the item name to be modified
        :param item_value: value of the item
        """
        if not self.is_data_name_existing(item_name):
            raise ValueError("Item {} not in grammar {}".format(item_name, self.name))
        self_dict = self.schema.to_schema()
        self_dict[self.PROPERTIES_FIELD][item_name] = item_value

        self.__set_grammar_from_dict(self_dict)

    def write_schema(self, path=None):
        """Write the schema to a file.

        Args:
            path: Write to this path, if None then write to a file named after the
                grammar and with .json extension.
        """
        if path is None:
            path = Path(self.name).with_suffix(".json")
        else:
            path = Path(path)

        schema_json = self.schema.to_json()

        if PY2:
            # workaround, see https://stackoverflow.com/a/36003774
            x = json.dumps(
                schema_json,
                ensure_ascii=False,
            )
            if isinstance(x, str):
                x = unicode(x, "UTF-8")  # noqa: F821
            path.write_text(x)
        else:
            json.dump(
                schema_json,
                path.open("w", encoding="utf-8"),
            )

    def __getstate__(self):
        """Used by pickle to define what to serialize.

        :returns: the dict to serialize
        """
        out_d = dict(self.__dict__)
        out_d.pop("_validator")
        # genson schema cannot be pickled: use its dictionary representation
        out_d["schema"] = self.schema.to_schema()
        return out_d

    def __setstate__(self, data_dict):
        """Used by pickle to define what to deserialize.

        :param data_dict : update self dict from data_dict to deserialize
        """
        self.__dict__.update(data_dict)
        # genson schema cannot be pickled: use its dictionary representation
        self.__set_grammar_from_dict(data_dict.pop("schema"))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A quadratic analytical problem
******************************
"""
from __future__ import division, unicode_literals

import logging

from numpy import array
from numpy import sum as np_sum

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction

LOGGER = logging.getLogger(__name__)


class Power2(OptimizationProblem):
    """**Power2** is a very basic quadratic analytical
    :class:`.OptimizationProblem`:

    - Objective to minimize: :math:`x_{dv,0}^2+x_1^2+x_2^2`
    - Inequality constraint 1: :math:`x_{dv,0}^3 - 0.5 > 0`
    - Inequality constraint 2: :math:`x_{dv,1}^3 - 0.5 > 0`
    - Equality constraint: :math:`x_{dv,2}^3 - 0.9 = 0`
    - Analytical optimum: :math:`(0.5^{1/3}, 0.5^{1/3}, 0.9^{1/3})`
    """

    def __init__(self, exception_error=False, initial_value=1.0):
        """The constructor initializes the Power2 :class:`.OptimizationProblem` by
        defining the :class:`.DesignSpace`, the objective function and the constraints.

        :param exception_error: if True, call to the objective raises errors
            useful for tests
        :type exception_error: bool
        """
        design_space = DesignSpace()
        design_space.add_variable("x", 3, l_b=-1.0, u_b=1.0, value=initial_value)

        super(Power2, self).__init__(design_space)
        self.objective = MDOFunction(
            self.pow2,
            name="pow2",
            f_type="obj",
            jac=self.pow2_jac,
            expr="x[0]**2+x[1]**2+x[2]**2",
            args=["x"],
        )
        ineq1 = MDOFunction(
            self.ineq_constraint1,
            name="ineq1",
            f_type="ineq",
            jac=self.ineq_constraint1_jac,
            expr="0.5 -x[0] ** 3",
            args=["x"],
        )
        self.add_ineq_constraint(ineq1)

        ineq2 = MDOFunction(
            self.ineq_constraint2,
            name="ineq2",
            f_type="ineq",
            jac=self.ineq_constraint2_jac,
            expr="0.5 -x[1] ** 3",
            args=["x"],
        )
        self.add_ineq_constraint(ineq2)

        eq_c = MDOFunction(
            self.eq_constraint,
            name="eq",
            f_type="eq",
            jac=self.eq_constraint_jac,
            expr="x[2] ** 3 - 0.9",
            args=["x"],
        )
        self.add_eq_constraint(eq_c)
        self.iter_error = 0
        self.exception_error = exception_error

    def pow2(self, x_dv):
        """Compute the objective of analytical function.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: objective value
        :rtype: float
        """
        if self.exception_error:
            if self.iter_error < 3:
                self.iter_error += 1
                obj = np_sum(x_dv ** 2)
            else:
                raise ValueError
        else:
            obj = np_sum(x_dv ** 2)
        return obj

    @staticmethod
    def pow2_jac(x_dv):
        """Compute the gradient of objective.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the objective gradient
        :rtype: numpy array
        """
        return 2 * x_dv

    @staticmethod
    def ineq_constraint1(x_dv):
        """Compute the first inequality constraint.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the first inequality constraint
        :rtype: numpy array
        """
        return -array([(x_dv[0] ** 3 - 0.5)])

    @staticmethod
    def ineq_constraint2(x_dv):
        """Compute the second inequality constraint.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the second inequality constraint
        :rtype: numpy array
        """
        return -array([(x_dv[1] ** 3 - 0.5)])

    @staticmethod
    def eq_constraint(x_dv):
        """Compute the equality constraint.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the equality constraint
        :rtype: numpy array
        """
        return -array([x_dv[2] ** 3 - 0.9])

    @staticmethod
    def ineq_constraint1_jac(x_dv):
        """Compute the first inequality constraint gradient.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the first inequality constraint gradient
        :rtype: numpy array
        """
        return -array([3 * x_dv[0] * x_dv[0], 0.0, 0.0])

    @staticmethod
    def ineq_constraint2_jac(x_dv):
        """Compute the second inequality constraint gradient.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the second inequality constraint gradient
        :rtype: numpy array
        """
        return -array([0, 3 * x_dv[1] * x_dv[1], 0.0])

    @staticmethod
    def eq_constraint_jac(x_dv):
        """Compute the equality constraint gradient.

        :param x_dv: design variable vector
        :type x_dv: numpy array
        :returns: value of the equality constraint gradient
        :rtype: numpy array
        """
        return -array([0.0, 0.0, 3 * x_dv[2] * x_dv[2]])

    @staticmethod
    def get_solution():
        """Return analytical result of optimization.

        :returns: theoretical optimum
        :rtype: numpy array
        """
        x_opt = array([0.5 ** (1.0 / 3.0), 0.5 ** (1.0 / 3.0), 0.9 ** (1.0 / 3.0)])
        f_opt = Power2().pow2(x_opt)
        return x_opt, f_opt



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The random forest for regression.

The random forest regression uses averaging methods on an ensemble of decision trees.

Dependence
----------
The regression model relies on the RandomForestRegressor class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.ensemble.RandomForestRegressor.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Iterable, Optional, Union

from numpy import ndarray
from sklearn.ensemble import RandomForestRegressor as SKLRandForest

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import TransformerType
from gemseo.mlearning.regression.regression import MLRegressionAlgo

LOGGER = logging.getLogger(__name__)


class RandomForestRegressor(MLRegressionAlgo):
    """Random forest regression."""

    LIBRARY = "scikit-learn"
    ABBR = "RandomForestRegressor"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        n_estimators=100,  # type: int
        **parameters
    ):  # type: (...) -> None
        """
        Args:
            n_estimators (int, optional): The number of trees in the forest.
        """
        super(RandomForestRegressor, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            n_estimators=n_estimators,
            **parameters  # type: Optional[Union[bool,int,float,str]]
        )
        self.algo = SKLRandForest(n_estimators=n_estimators, **parameters)

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        # SKLearn RandomForestReressor does not like output
        # shape (n_samples, 1), prefers (n_samples,).
        # The shape (n_samples, n_outputs) with n_outputs >= 2 is fine.
        if output_data.shape[1] == 1:
            output_data = output_data[:, 0]
        self.algo.fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        output_data = self.algo.predict(input_data)

        # n_outputs=1 => output_shape=(n_samples,). Convert to (n_samples, 1).
        if len(output_data.shape) == 1:
            output_data = output_data[:, None]

        return output_data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Provide routines for XDSM and tikz."""

from __future__ import division, unicode_literals

from pyxdsm.XDSM import XDSM

from gemseo.utils.py23_compat import Path, string_types


class XDSMToPDFConverter(object):
    """Convert an XDSM to a PDF file with tikz and latex."""

    def __init__(self):  # noqa: D107
        self.__xdsm = XDSM()

    def convert(self, xdsm_data, directory_path, filename_without_ext, scenario):
        """Convert a dictionary representation of a XDSM into a pdf.

        Args:
            xdsm_data: XDSM dictionary representation.
            directory_path (str): output directory.
            filename_without_ext (str): output file name, default is 'xdsm'.
            scenario (str): scenario name, default is 'root'.
            quiet (bool): set to True to suppress output from pdflatex.
        """
        workflow = xdsm_data[scenario]["workflow"][1]
        numbers = {}

        self.__get_numbers(numbers, workflow)
        self.__add_nodes(numbers, xdsm_data[scenario]["nodes"])
        self.__add_edges(xdsm_data[scenario]["edges"])
        self.__add_processes(workflow)

        # workaround xdsm not well handling file path: latex expects posix path
        file_path = str(directory_path / filename_without_ext).replace("\\", "/")

        self.__xdsm.write(file_path)

        if not Path(filename_without_ext).with_suffix(".pdf").exists():
            raise RuntimeError(
                "Something went wrong during the Latex compilation,"
                " as xdsm.pdf has not been generated. Please have a look at the"
                " Latex log file to investigate the root cause of the error."
            )

        # Build XDSM for sub-scenarios
        if scenario == "root":
            subscenarios = [key for key in xdsm_data.keys() if "scn-" in key]
            for subscenario in subscenarios:
                self.convert(
                    xdsm_data,
                    directory_path,
                    filename_without_ext=filename_without_ext + "_" + subscenario,
                    scenario=subscenario,
                )

    def __add_processes(self, workflow, prev=None):
        """Create the pyXDSM processes from a given workflow in a recursive way.

        Args:
            workflow (list): workflow component of the dictionary storing the XDSM.
            prev (str): name of the previous node.
        """
        systems = []
        last_node = None
        for idx, system in enumerate(workflow):
            if isinstance(system, string_types):
                # system is a node
                systems.append(system)
                last_node = system
            elif isinstance(system, list):
                # system is a group of nodes
                if isinstance(system[0], string_types):
                    # system[0] is a node
                    if last_node is not None:
                        self.__xdsm.add_process([last_node, system[0]], arrow=True)
                    self.__add_processes(system, workflow[idx - 1])
                elif isinstance(system[0], dict):
                    # system[0] is a parallel sequence
                    for sub_sys in system[0]["parallel"]:
                        # add a sub-process for each parallel element
                        sub_workflow = [last_node, sub_sys]
                        self.__add_processes(sub_workflow, last_node)
        if prev is not None:
            systems.append(prev)
        self.__xdsm.add_process(systems, arrow=True)

    def __get_numbers(self, numbers, nodes, current=0, end=0, following=1):
        """Give number to the different nodes in a recursive way.

        Args:
            numbers: dictionary.
            nodes (list): workflow component of the dictionary storing the XDSM.
            current (int): current step index.
            end (int): end-loop step index.
            following (int): following step index.

        Returns:
            The following step index.
        """
        prev_node = "undefined"
        for node in nodes:
            if isinstance(node, string_types):
                current = following
                following = current + 1
                end = current
                current_l = [current]
                if node in list(numbers.keys()):
                    current_l = numbers[node]["current"] + current_l
                numbers[node] = {"current": current_l, "next": following, "end": end}
                prev_node = node
            elif isinstance(node, list):
                following = self.__get_numbers(numbers, node, current, end, following)
                numbers[prev_node]["end"] = following
                following += 1
            elif isinstance(node, dict):
                for sub_node in node["parallel"]:
                    if not isinstance(sub_node, list):
                        following = self.__get_numbers(
                            numbers, [sub_node], current, end, following
                        )
                    else:
                        following = self.__get_numbers(
                            numbers, sub_node, current, end, following
                        )
        return following

    def __add_nodes(self, numbers, nodes):
        """Add the different nodes, called 'systems', in the XDSM."""

        for node in nodes:
            name = ",".join(
                [str(current) for current in numbers[node["id"]]["current"]]
            )

            name_1 = name + ",{}-{}:".format(
                str(numbers[node["id"]]["end"]), str(numbers[node["id"]]["next"])
            )
            name_2 = name + ":"

            if node["type"] == "optimization":
                node_type = "Optimization"
                name = name_1
            elif node["type"] == "lp_optimization":
                node_type = "LP_Optimization"
                name = name_1
            if node["type"] == "doe":
                node_type = "DOE"
                name = name_1
            elif node["type"] == "mda":
                node_type = "MDA"
                name = name_1
            elif node["type"] == "mdo":
                node_type = "MDO"
                name = name_2
            elif node["type"] == "analysis":
                node_type = "Function"
                name = name_2
            elif node["type"] == "function":
                node_type = "Function"
                name = name_2
            elif node["type"] == "metamodel":
                node_type = "Metamodel"
                name = name_2

            node_replaced = node["name"]
            escaped_characters = ["_", "$", "&", "{", "}", "%"]
            for char in escaped_characters:
                node_replaced = node_replaced.replace(char, r"\{}".format(char))
            name = name + node_replaced

            self.__xdsm.add_system(node["id"], node_type, r"\text{" + name + "}")

    def __add_edges(self, edges):
        """Add the edges called connections, inputs, outputs to the XDSM."""

        for edge in edges:
            old_names = edge["name"].split(",")
            names = []

            for name in old_names:
                sub = sup = None
                s_name = name.split("^")
                if len(s_name) == 2:
                    name = s_name[0]
                    suffix = s_name[1]
                    s_name = suffix.split("_")
                    if len(s_name) == 1:
                        s_name = name.split("_")
                        if len(s_name) > 1:
                            sup = suffix
                            sub = s_name[-1]
                            name = "".join(s_name[:-1])
                    else:
                        sub = s_name[0]
                        sup = s_name[1]
                else:
                    s_name = name.split("_")
                    if len(s_name) > 1:
                        sub = s_name[-1]
                        name = "".join(s_name[:-1])

                sup = "" if sup is None else "^{" + sup + "}"
                sub = "" if sub is None else "_{" + sub + "}"
                names.append(name + sub + sup)

            names = ", ".join(names)

            if edge["to"] == "_U_":
                self.__xdsm.add_output(edge["from"], r"" + names + "", side="left")
            elif edge["from"] != "_U_":
                self.__xdsm.connect(edge["from"], edge["to"], r"" + names + "")
            elif edge["from"] == "_U_":
                self.__xdsm.add_input("Opt", r"" + names + "")


def xdsm_data_to_pdf(
    xdsm_data, directory_path, filename_without_ext="xdsm", scenario="root"
):
    """Convert a dictionary representation of a XDSM to a pdf.

    Args:
        xdsm_data: XDSM dictionary representation.
        directory_path (str): output directory.
        filename_without_ext (str): output file name, default is 'xdsm'.
        scenario (str): scenario name, default is 'root'.
    """
    converter = XDSMToPDFConverter()
    converter.convert(xdsm_data, directory_path, filename_without_ext, scenario)
